from typing import Mapping, Optional, Set, Deque, Sequence
from collections import deque, defaultdict
from yellowstone_fumarole_proto.fumarole_pb2 import (
    CommitmentLevel,
    BlockchainEvent,
)
from yellowstone_fumarole_client.utils.collections import OrderedSet
import heapq
from enum import Enum

__all__ = [
    "DEFAULT_SLOT_MEMORY_RETENTION",
    "FumeBlockchainId",
    "FumeBlockUID",
    "FumeNumShards",
    "FumeShardIdx",
    "FumeOffset",
    "FumeSessionSequence",
    "Slot",
    "FumeDownloadRequest",
    "FumeSlotStatus",
    "SlotCommitmentProgression",
    "SlotDownloadProgress",
    "SlotDownloadState",
    "FumaroleSM",
]

# Constants
DEFAULT_SLOT_MEMORY_RETENTION = 10000

# Type aliases
FumeBlockchainId = bytes  # Equivalent to [u8; 16]
FumeBlockUID = bytes  # Equivalent to [u8; 16]
FumeNumShards = int  # Equivalent to u32
FumeShardIdx = int  # Equivalent to u32
FumeOffset = int  # Equivalent to i64
FumeSessionSequence = int  # Equivalent to u64
Slot = int  # From solana_sdk::clock::Slot


# Data structures
class FumeDownloadRequest:
    def __init__(
        self,
        slot: Slot,
        blockchain_id: FumeBlockchainId,
        block_uid: FumeBlockUID,
        num_shards: FumeNumShards,
        commitment_level: CommitmentLevel,
    ):
        self.slot = slot
        self.blockchain_id = blockchain_id
        self.block_uid = block_uid
        self.num_shards = num_shards
        self.commitment_level = commitment_level


class FumeSlotStatus:
    def __init__(
        self,
        session_sequence: FumeSessionSequence,
        offset: FumeOffset,
        slot: Slot,
        parent_slot: Optional[Slot],
        commitment_level: CommitmentLevel,
        dead_error: Optional[str],
    ):
        self.session_sequence = session_sequence
        self.offset = offset
        self.slot = slot
        self.parent_slot = parent_slot
        self.commitment_level = commitment_level
        self.dead_error = dead_error


class SlotCommitmentProgression:
    def __init__(self):
        self.processed_commitment_levels: Set[CommitmentLevel] = set()


class SlotDownloadProgress:
    def __init__(self, num_shards: FumeNumShards):
        self.num_shards = num_shards
        self.shard_remaining = [False] * num_shards

    def do_progress(self, shard_idx: FumeShardIdx) -> "SlotDownloadState":
        self.shard_remaining[shard_idx % self.num_shards] = True
        return (
            SlotDownloadState.Done
            if all(self.shard_remaining)
            else SlotDownloadState.Downloading
        )


class SlotDownloadState(Enum):
    Downloading = "Downloading"
    Done = "Done"


class FumaroleSM:
    """
    Sans-IO Fumarole State Machine

    Manages in-flight slot downloads and ensures correct ordering of slot statuses without performing I/O.
    """

    def __init__(self, last_committed_offset: FumeOffset, slot_memory_retention: int):
        self.last_committed_offset = last_committed_offset
        self.slot_commitment_progression: dict[Slot, SlotCommitmentProgression] = dict()  # Slot -> SlotCommitmentProgression
        self.downloaded_slot = OrderedSet()  # Set of downloaded slots
        self.inflight_slot_shard_download = {}  # Slot -> SlotDownloadProgress
        self.blocked_slot_status_update = defaultdict(
            deque
        )  # Slot -> Deque[FumeSlotStatus]
        self.slot_status_update_queue: Deque[FumeSlotStatus] = deque()  # Deque[FumeSlotStatus]
        self.processed_offset = []  # Min-heap for (sequence, offset)
        self.committable_offset = last_committed_offset
        self.max_slot_detected = 0
        self.unprocessed_blockchain_event: Deque[
            (FumeSessionSequence, BlockchainEvent)
        ] = deque()
        self.sequence = 1
        self.last_processed_fume_sequence = 0
        self.sequence_to_offset: dict[FumeSessionSequence, FumeOffset] = (
            {}
        )  # FumeSessionSequence -> FumeOffset
        self.slot_memory_retention = slot_memory_retention

    def update_committed_offset(self, offset: FumeOffset) -> None:
        assert (
            offset >= self.last_committed_offset
        ), "Offset must be >= last committed offset"
        self.last_committed_offset = offset

    def next_sequence(self) -> int:
        ret = self.sequence
        self.sequence += 1
        return ret

    def gc(self) -> None:
        """Garbage collect old slots to respect memory retention limit."""
        while len(self.downloaded_slot) > self.slot_memory_retention:
            slot = self.downloaded_slot.popfirst() if self.downloaded_slot else None
            if slot is None:
                break
            self.slot_commitment_progression.pop(slot, None)
            self.inflight_slot_shard_download.pop(slot, None)
            self.blocked_slot_status_update.pop(slot, None)

    def queue_blockchain_event(self, events: Sequence[BlockchainEvent]) -> None:
        """Queue blockchain events for processing."""
        for event in events:

            if event.offset < self.last_committed_offset:
                continue

            if event.slot > self.max_slot_detected:
                self.max_slot_detected = event.slot

            sequence = self.next_sequence()
            self.sequence_to_offset[sequence] = event.offset

            if event.slot in self.downloaded_slot:
                fume_status = FumeSlotStatus(
                    session_sequence=sequence,
                    offset=event.offset,
                    slot=event.slot,
                    parent_slot=event.parent_slot,
                    commitment_level=event.commitment_level,
                    dead_error=event.dead_error,
                )

                if event.slot in self.inflight_slot_shard_download:
                    self.blocked_slot_status_update[event.slot].append(fume_status)
                else:
                    self.slot_status_update_queue.append(fume_status)
            else:
                self.unprocessed_blockchain_event.append((sequence, event))

    def make_slot_download_progress(
        self, slot: Slot, shard_idx: FumeShardIdx
    ) -> SlotDownloadState:
        """Update download progress for a given slot."""
        download_progress = self.inflight_slot_shard_download.get(slot)
        if not download_progress:
            raise ValueError("Slot not in download")

        download_state = download_progress.do_progress(shard_idx)

        if download_state == SlotDownloadState.Done:
            self.inflight_slot_shard_download.pop(slot)
            self.downloaded_slot.add(slot)
            self.slot_commitment_progression.setdefault(
                slot, SlotCommitmentProgression()
            )
            blocked_statuses = self.blocked_slot_status_update.pop(slot, deque())
            self.slot_status_update_queue.extend(blocked_statuses)

        return download_state

    def pop_next_slot_status(self) -> Optional[FumeSlotStatus]:
        """Pop the next slot status to process."""
        while self.slot_status_update_queue:
            slot_status = self.slot_status_update_queue.popleft()
            commitment_history = self.slot_commitment_progression.get(slot_status.slot)
            if commitment_history is None:
                raise RuntimeError("Slot status should not be available here")
            if slot_status.commitment_level not in commitment_history.processed_commitment_levels:
                commitment_history.processed_commitment_levels.add(
                    slot_status.commitment_level
                )
                return slot_status
            else:
                # Already processed this commitment level
                self.mark_event_as_processed(slot_status.session_sequence)
        return None

    def make_sure_slot_commitment_progression_exists(
        self, slot: Slot
    ) -> SlotCommitmentProgression:
        """Ensure a slot has a commitment progression entry."""
        return self.slot_commitment_progression.setdefault(
            slot, SlotCommitmentProgression()
        )

    def pop_slot_to_download(self, commitment=None) -> Optional[FumeDownloadRequest]:
        """Pop the next slot to download."""
        min_commitment = commitment or CommitmentLevel.PROCESSED
        while self.unprocessed_blockchain_event:
            session_sequence, blockchain_event = (
                self.unprocessed_blockchain_event.popleft()
            )
            blockchain_event: BlockchainEvent = blockchain_event
            event_cl = blockchain_event.commitment_level

            if event_cl != min_commitment:
                self.slot_status_update_queue.append(
                    FumeSlotStatus(
                        session_sequence=session_sequence,
                        offset=blockchain_event.offset,
                        slot=blockchain_event.slot,
                        parent_slot=blockchain_event.parent_slot,
                        commitment_level=event_cl,
                        dead_error=blockchain_event.dead_error,
                    )
                )
                self.make_sure_slot_commitment_progression_exists(blockchain_event.slot)
                continue

            if blockchain_event.slot in self.downloaded_slot:
                self.make_sure_slot_commitment_progression_exists(blockchain_event.slot)
                progression = self.slot_commitment_progression[blockchain_event.slot]
                if event_cl in progression.processed_commitment_levels:
                    self.mark_event_as_processed(session_sequence)
                    continue

                self.slot_status_update_queue.append(
                    FumeSlotStatus(
                        session_sequence=session_sequence,
                        offset=blockchain_event.offset,
                        slot=blockchain_event.slot,
                        parent_slot=blockchain_event.parent_slot,
                        commitment_level=event_cl,
                        dead_error=blockchain_event.dead_error,
                    )
                )
            else:
                blockchain_id = bytes(blockchain_event.blockchain_id)
                block_uid = bytes(blockchain_event.block_uid)

                self.blocked_slot_status_update[blockchain_event.slot].append(
                    FumeSlotStatus(
                        session_sequence=session_sequence,
                        offset=blockchain_event.offset,
                        slot=blockchain_event.slot,
                        parent_slot=blockchain_event.parent_slot,
                        commitment_level=event_cl,
                        dead_error=blockchain_event.dead_error,
                    )
                )

                if blockchain_event.slot not in self.inflight_slot_shard_download:
                    download_request = FumeDownloadRequest(
                        slot=blockchain_event.slot,
                        blockchain_id=blockchain_id,
                        block_uid=block_uid,
                        num_shards=blockchain_event.num_shards,
                        commitment_level=event_cl,
                    )
                    download_progress = SlotDownloadProgress(
                        num_shards=blockchain_event.num_shards
                    )
                    self.inflight_slot_shard_download[blockchain_event.slot] = (
                        download_progress
                    )
                    return download_request
        return None

    def mark_event_as_processed(self, event_seq_number: FumeSessionSequence) -> None:
        """Mark an event as processed and update committable offset."""
        fume_offset = self.sequence_to_offset.pop(event_seq_number, None)
        if fume_offset is None:
            raise ValueError("Event sequence number not found")
        heapq.heappush(self.processed_offset, (event_seq_number, fume_offset))

        while self.processed_offset:
            seq, offset = self.processed_offset[0]
            if seq != self.last_processed_fume_sequence + 1:
                break
            heapq.heappop(self.processed_offset)
            self.committable_offset = offset
            self.last_processed_fume_sequence = seq

    def slot_status_update_queue_len(self) -> int:
        """Return the length of the slot status update queue."""
        return len(self.slot_status_update_queue)

    def processed_offset_queue_len(self) -> int:
        """Return the length of the processed offset queue."""
        return len(self.processed_offset)

    def need_new_blockchain_events(self) -> bool:
        """Check if new blockchain events are needed."""
        MINIMUM_UNPROCESSED_BLOCKCHAIN_EVENT = 10
        return len(self.unprocessed_blockchain_event) < MINIMUM_UNPROCESSED_BLOCKCHAIN_EVENT \
            or (not self.slot_status_update_queue and not self.blocked_slot_status_update)



import asyncio
import logging
from contextlib import suppress

from yellowstone_fumarole_client.grpc_connectivity import (
    FumaroleGrpcConnector,
)
from typing import AsyncGenerator, Optional
from dataclasses import dataclass
from yellowstone_fumarole_client.config import FumaroleConfig
from yellowstone_fumarole_client.runtime.aio import (
    AsyncioFumeDragonsmouthRuntime,
    DEFAULT_GC_INTERVAL,
    DEFAULT_SLOT_MEMORY_RETENTION,
    GrpcSlotDownloader,
)
from yellowstone_fumarole_client.runtime.state_machine import (
    FumaroleSM,
    FumeOffset,
)
from yellowstone_fumarole_proto.geyser_pb2 import SubscribeRequest, SubscribeUpdate
from yellowstone_fumarole_proto.fumarole_pb2 import (
    ControlResponse,
    VersionRequest,
    VersionResponse,
    JoinControlPlane,
    ControlCommand,
    ListConsumerGroupsRequest,
    ListConsumerGroupsResponse,
    GetConsumerGroupInfoRequest,
    ConsumerGroupInfo,
    DeleteConsumerGroupRequest,
    DeleteConsumerGroupResponse,
    CreateConsumerGroupRequest,
    CreateConsumerGroupResponse,
)
from yellowstone_fumarole_proto.fumarole_pb2_grpc import FumaroleStub
from yellowstone_fumarole_client.error import SubscribeError
import grpc

from yellowstone_fumarole_client import config

__all__ = [
    "FumaroleClient",
    "FumaroleConfig",
    "FumaroleSubscribeConfig",
    "DragonsmouthAdapterSession",
    "DEFAULT_DRAGONSMOUTH_CAPACITY",
    "DEFAULT_COMMIT_INTERVAL",
    "DEFAULT_MAX_SLOT_DOWNLOAD_ATTEMPT",
    "DEFAULT_CONCURRENT_DOWNLOAD_LIMIT_PER_TCP",
]

# Constants
DEFAULT_DRAGONSMOUTH_CAPACITY = 100000
DEFAULT_COMMIT_INTERVAL = 5.0  # seconds
DEFAULT_MAX_SLOT_DOWNLOAD_ATTEMPT = 3
DEFAULT_CONCURRENT_DOWNLOAD_LIMIT_PER_TCP = 10

LOGGER = logging.getLogger(__name__)
# Error classes


# FumaroleSubscribeConfig
@dataclass
class FumaroleSubscribeConfig:
    """Configuration for subscribing to a dragonsmouth stream."""

    # The maximum number of concurrent download tasks per TCP connection.
    concurrent_download_limit: int = DEFAULT_CONCURRENT_DOWNLOAD_LIMIT_PER_TCP

    # The interval at which to commit the slot memory.
    commit_interval: float = DEFAULT_COMMIT_INTERVAL

    # The maximum number of failed slot download attempts before giving up.
    max_failed_slot_download_attempt: int = DEFAULT_MAX_SLOT_DOWNLOAD_ATTEMPT

    # The maximum number of slots to download concurrently.
    data_channel_capacity: int = DEFAULT_DRAGONSMOUTH_CAPACITY

    # The interval at which to perform garbage collection on the slot memory.
    gc_interval: int = DEFAULT_GC_INTERVAL

    # How many processed slot numbers to retain in memory to avoid duplication.
    slot_memory_retention: int = DEFAULT_SLOT_MEMORY_RETENTION


@dataclass
class FumaroleSubscribeStats:
    """Commit/slot statistics for the Fumarole subscribe session."""

    # Last committed log offset in Fumarole -- this is a low-level, implementation detail.
    # NOTE: this should not be part as business logic, can change any time.
    log_committed_offset: FumeOffset
    # NOTE:: this is a low-level information, can change any time.
    log_committable_offset: FumeOffset
    # Max slot seen by the in the current session - does not mean it has been processed.
    max_slot_seen: int
    # Number of committable offsets blocked by an ancestor commit offset not yet processed.
    processed_offset_queue_len: int = 0  


# DragonsmouthAdapterSession
@dataclass
class DragonsmouthAdapterSession:
    """Session for interacting with the dragonsmouth-like stream."""

    # The queue for sending SubscribeRequest update to the dragonsmouth stream.
    sink: asyncio.Queue

    # The queue for receiving SubscribeUpdate from the dragonsmouth stream.
    source: AsyncGenerator[SubscribeUpdate, None]

    # The task handle for the fumarole runtime.
    _fumarole_handle: asyncio.Task

    _sm: FumaroleSM

    async def __aenter__(self):
        """Enter the session context."""
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        self.sink.shutdown()
        self._fumarole_handle.cancel()

    def stats(self) -> FumaroleSubscribeStats:
        """Get low-level statistics of the Fumarole state-machine."""
        commitable = self._sm.committable_offset
        committed = self._sm.last_committed_offset
        max_slot = self._sm.max_slot_detected
        processed_offset_queue_len = len(self._sm.processed_offset)
        return FumaroleSubscribeStats(
            log_committed_offset=committed,
            log_committable_offset=commitable,
            max_slot_seen=max_slot,
            processed_offset_queue_len=processed_offset_queue_len,
        )


# FumaroleClient
class FumaroleClient:
    """Fumarole client for interacting with the Fumarole server."""

    logger = logging.getLogger(__name__)

    def __init__(self, connector: FumaroleGrpcConnector, stub: FumaroleStub):
        self.connector = connector
        self.stub = stub

    @staticmethod
    async def connect(config: config.FumaroleConfig) -> "FumaroleClient":
        """Connect to the Fumarole server using the provided configuration.
        Args:
            config (FumaroleConfig): Configuration for the Fumarole client.
        """
        endpoint = config.endpoint
        connector = FumaroleGrpcConnector(config=config, endpoint=endpoint)
        FumaroleClient.logger.debug(f"Connecting to {endpoint}")
        client = await connector.connect()
        FumaroleClient.logger.debug(f"Connected to {endpoint}")
        return FumaroleClient(connector=connector, stub=client)

    async def version(self) -> VersionResponse:
        """Get the version of the Fumarole server."""
        request = VersionRequest()
        response = await self.stub.Version(request)
        return response

    async def dragonsmouth_subscribe(
        self, consumer_group_name: str, request: SubscribeRequest
    ) -> DragonsmouthAdapterSession:
        """Subscribe to a dragonsmouth stream with default configuration.

        Args:
            consumer_group_name (str): The name of the consumer group.
            request (SubscribeRequest): The request to subscribe to the dragonsmouth stream.
        """
        return await self.dragonsmouth_subscribe_with_config(
            consumer_group_name, request, FumaroleSubscribeConfig()
        )

    async def dragonsmouth_subscribe_with_config(
        self,
        consumer_group_name: str,
        request: SubscribeRequest,
        config: FumaroleSubscribeConfig,
    ) -> DragonsmouthAdapterSession:
        """Subscribe to a dragonsmouth stream with custom configuration.

        Args:
            consumer_group_name (str): The name of the consumer group.
            request (SubscribeRequest): The request to subscribe to the dragonsmouth stream.
            config (FumaroleSubscribeConfig): The configuration for the dragonsmouth subscription.
        """
        dragonsmouth_outlet = asyncio.Queue(maxsize=config.data_channel_capacity)
        fume_control_plane_q = asyncio.Queue(maxsize=100)

        initial_join = JoinControlPlane(consumer_group_name=consumer_group_name)
        initial_join_command = ControlCommand(initial_join=initial_join)
        await fume_control_plane_q.put(initial_join_command)

        FumaroleClient.logger.debug(
            f"Sent initial join command: {initial_join_command}"
        )

        async def control_plane_sink():
            while True:
                try:
                    update = await fume_control_plane_q.get()
                    yield update
                except (asyncio.CancelledError, asyncio.QueueShutDown):
                    break

        fume_control_plane_stream_rx: grpc.aio.StreamStreamCall = self.stub.Subscribe(
            control_plane_sink()
        )  # it's actually InterceptedStreamStreamCall, but grpc lib doesn't export it

        control_response: ControlResponse = await fume_control_plane_stream_rx.read()
        init = control_response.init
        if init is None:
            raise ValueError(f"Unexpected initial response: {control_response}")

        # Once we have the initial response, we can spin a task to read from the stream
        # and put the updates into the queue.
        # This is a bit of a hack, but we need a Queue not a StreamStreamMultiCallable
        # because Queue are cancel-safe, while Stream are not, or at least didn't find any docs about it.
        fume_control_plane_rx_q = asyncio.Queue(maxsize=100)

        async def control_plane_source():
            while True:
                try:
                    async for update in fume_control_plane_stream_rx:
                        await fume_control_plane_rx_q.put(update)
                except asyncio.QueueShutDown:
                    break
                except asyncio.CancelledError:
                    break
                finally:
                    fume_control_plane_rx_q.shutdown()

        control_plane_src_task = asyncio.create_task(control_plane_source())

        FumaroleClient.logger.debug(f"Control response: {control_response}")

        last_committed_offset = init.last_committed_offsets.get(0)
        if last_committed_offset is None:
            raise ValueError("No last committed offset")

        sm = FumaroleSM(last_committed_offset, config.slot_memory_retention)
        subscribe_request_queue = asyncio.Queue(maxsize=100)

        data_plane_client = await self.connector.connect()

        grpc_slot_downloader = GrpcSlotDownloader(
            client=data_plane_client,
        )

        rt = AsyncioFumeDragonsmouthRuntime(
            sm=sm,
            slot_downloader=grpc_slot_downloader,
            subscribe_request_update_q=subscribe_request_queue,
            subscribe_request=request,
            consumer_group_name=consumer_group_name,
            control_plane_tx_q=fume_control_plane_q,
            control_plane_rx_q=fume_control_plane_rx_q,
            dragonsmouth_outlet=dragonsmouth_outlet,
            commit_interval=config.commit_interval,
            gc_interval=config.gc_interval,
            max_concurrent_download=config.concurrent_download_limit,
        )

        rt_task = asyncio.create_task(rt.run())
        rt_task.set_name("rt_task")
        control_plane_src_task.set_name("control_plane_src_task")
        async def fumarole_overseer():
            try:
                done, pending = await asyncio.wait(
                    [rt_task, control_plane_src_task], return_when=asyncio.FIRST_COMPLETED
                )
                for t in done:
                    try:
                        exc = t.exception()
                        LOGGER.error(f"Fumarole task '{t.get_name()}' failed: {exc}")
                    except asyncio.CancelledError:
                        pass
                    if exc is not None:
                        with suppress(asyncio.QueueShutDown):
                            await dragonsmouth_outlet.put(exc)
                for t in pending:
                    t.cancel()
            finally:
                await rt.aclose()


        fumarole_handle = asyncio.create_task(fumarole_overseer())

        async def source_gen() -> AsyncGenerator[SubscribeUpdate, None]:
            try:
                while True:
                    update = await dragonsmouth_outlet.get()
                    if isinstance(update, Exception):
                        raise update
                    yield update
            except (asyncio.CancelledError, asyncio.QueueShutDown):
                pass
            finally:
                dragonsmouth_outlet.shutdown()

        return DragonsmouthAdapterSession(
            sink=subscribe_request_queue,
            source=source_gen(),
            _fumarole_handle=fumarole_handle,
            _sm=sm,
        )

    async def list_consumer_groups(
        self,
    ) -> ListConsumerGroupsResponse:
        """Lists all consumer groups."""
        return await self.stub.ListConsumerGroups(ListConsumerGroupsRequest())

    async def get_consumer_group_info(
        self, consumer_group_name: str
    ) -> Optional[ConsumerGroupInfo]:
        """Gets information about a consumer group by name.
        Returns None if the consumer group does not exist.

        Args:
            consumer_group_name (str): The name of the consumer group to retrieve information for.
        """
        try:
            return await self.stub.GetConsumerGroupInfo(
                GetConsumerGroupInfoRequest(consumer_group_name=consumer_group_name)
            )
        except grpc.aio.AioRpcError as e:
            if e.code() == grpc.StatusCode.NOT_FOUND:
                return None
            else:
                raise

    async def delete_consumer_group(
        self, consumer_group_name: str
    ) -> DeleteConsumerGroupResponse:
        """Delete a consumer group by name.

        NOTE: this operation is idempotent, meaning that if the consumer group does not exist, it will not raise an error.
        Args:
            consumer_group_name (str): The name of the consumer group to delete.
        """
        return await self.stub.DeleteConsumerGroup(
            DeleteConsumerGroupRequest(consumer_group_name=consumer_group_name)
        )

    async def delete_all_consumer_groups(
        self,
    ) -> DeleteConsumerGroupResponse:
        """Deletes all consumer groups."""
        consumer_group_list = await self.list_consumer_groups()

        tasks = []

        async with asyncio.TaskGroup() as tg:
            for group in consumer_group_list.consumer_groups:
                cg_name = group.consumer_group_name
                task = tg.create_task(self.delete_consumer_group(cg_name))
                tasks.append((cg_name, task))

        # Raise an error if any task fails
        for cg_name, task in tasks:
            result = task.result()
            if not result.success:
                raise RuntimeError(
                    f"Failed to delete consumer group {cg_name}: {result.error}"
                )

    async def create_consumer_group(
        self, request: CreateConsumerGroupRequest
    ) -> CreateConsumerGroupResponse:
        """Creates a new consumer group.
        Args:
            request (CreateConsumerGroupRequest): The request to create a consumer group.
        """
        return await self.stub.CreateConsumerGroup(request)



import asyncio
import logging

LOGGER = logging.getLogger(__name__)


async def never():
    """
    Create a forever pending future. This future is not set and will never be set.
    This is useful for testing purposes.
    """
    loop = asyncio.get_running_loop()
    return await loop.create_future()


class Interval:

    def __init__(self, interval: float):
        """
        Create an interval that will run the given factory every `interval` seconds.

        Args:
            interval: The interval in seconds.
        """
        self.interval = interval

    async def tick(self):
        await asyncio.sleep(self.interval)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

import yellowstone_fumarole_proto.geyser_pb2 as geyser__pb2

GRPC_GENERATED_VERSION = '1.74.0'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in geyser_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class GeyserStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Subscribe = channel.stream_stream(
                '/geyser.Geyser/Subscribe',
                request_serializer=geyser__pb2.SubscribeRequest.SerializeToString,
                response_deserializer=geyser__pb2.SubscribeUpdate.FromString,
                _registered_method=True)
        self.SubscribeReplayInfo = channel.unary_unary(
                '/geyser.Geyser/SubscribeReplayInfo',
                request_serializer=geyser__pb2.SubscribeReplayInfoRequest.SerializeToString,
                response_deserializer=geyser__pb2.SubscribeReplayInfoResponse.FromString,
                _registered_method=True)
        self.Ping = channel.unary_unary(
                '/geyser.Geyser/Ping',
                request_serializer=geyser__pb2.PingRequest.SerializeToString,
                response_deserializer=geyser__pb2.PongResponse.FromString,
                _registered_method=True)
        self.GetLatestBlockhash = channel.unary_unary(
                '/geyser.Geyser/GetLatestBlockhash',
                request_serializer=geyser__pb2.GetLatestBlockhashRequest.SerializeToString,
                response_deserializer=geyser__pb2.GetLatestBlockhashResponse.FromString,
                _registered_method=True)
        self.GetBlockHeight = channel.unary_unary(
                '/geyser.Geyser/GetBlockHeight',
                request_serializer=geyser__pb2.GetBlockHeightRequest.SerializeToString,
                response_deserializer=geyser__pb2.GetBlockHeightResponse.FromString,
                _registered_method=True)
        self.GetSlot = channel.unary_unary(
                '/geyser.Geyser/GetSlot',
                request_serializer=geyser__pb2.GetSlotRequest.SerializeToString,
                response_deserializer=geyser__pb2.GetSlotResponse.FromString,
                _registered_method=True)
        self.IsBlockhashValid = channel.unary_unary(
                '/geyser.Geyser/IsBlockhashValid',
                request_serializer=geyser__pb2.IsBlockhashValidRequest.SerializeToString,
                response_deserializer=geyser__pb2.IsBlockhashValidResponse.FromString,
                _registered_method=True)
        self.GetVersion = channel.unary_unary(
                '/geyser.Geyser/GetVersion',
                request_serializer=geyser__pb2.GetVersionRequest.SerializeToString,
                response_deserializer=geyser__pb2.GetVersionResponse.FromString,
                _registered_method=True)


class GeyserServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Subscribe(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SubscribeReplayInfo(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Ping(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetLatestBlockhash(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetBlockHeight(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetSlot(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def IsBlockhashValid(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetVersion(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_GeyserServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Subscribe': grpc.stream_stream_rpc_method_handler(
                    servicer.Subscribe,
                    request_deserializer=geyser__pb2.SubscribeRequest.FromString,
                    response_serializer=geyser__pb2.SubscribeUpdate.SerializeToString,
            ),
            'SubscribeReplayInfo': grpc.unary_unary_rpc_method_handler(
                    servicer.SubscribeReplayInfo,
                    request_deserializer=geyser__pb2.SubscribeReplayInfoRequest.FromString,
                    response_serializer=geyser__pb2.SubscribeReplayInfoResponse.SerializeToString,
            ),
            'Ping': grpc.unary_unary_rpc_method_handler(
                    servicer.Ping,
                    request_deserializer=geyser__pb2.PingRequest.FromString,
                    response_serializer=geyser__pb2.PongResponse.SerializeToString,
            ),
            'GetLatestBlockhash': grpc.unary_unary_rpc_method_handler(
                    servicer.GetLatestBlockhash,
                    request_deserializer=geyser__pb2.GetLatestBlockhashRequest.FromString,
                    response_serializer=geyser__pb2.GetLatestBlockhashResponse.SerializeToString,
            ),
            'GetBlockHeight': grpc.unary_unary_rpc_method_handler(
                    servicer.GetBlockHeight,
                    request_deserializer=geyser__pb2.GetBlockHeightRequest.FromString,
                    response_serializer=geyser__pb2.GetBlockHeightResponse.SerializeToString,
            ),
            'GetSlot': grpc.unary_unary_rpc_method_handler(
                    servicer.GetSlot,
                    request_deserializer=geyser__pb2.GetSlotRequest.FromString,
                    response_serializer=geyser__pb2.GetSlotResponse.SerializeToString,
            ),
            'IsBlockhashValid': grpc.unary_unary_rpc_method_handler(
                    servicer.IsBlockhashValid,
                    request_deserializer=geyser__pb2.IsBlockhashValidRequest.FromString,
                    response_serializer=geyser__pb2.IsBlockhashValidResponse.SerializeToString,
            ),
            'GetVersion': grpc.unary_unary_rpc_method_handler(
                    servicer.GetVersion,
                    request_deserializer=geyser__pb2.GetVersionRequest.FromString,
                    response_serializer=geyser__pb2.GetVersionResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'geyser.Geyser', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('geyser.Geyser', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class Geyser(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Subscribe(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(
            request_iterator,
            target,
            '/geyser.Geyser/Subscribe',
            geyser__pb2.SubscribeRequest.SerializeToString,
            geyser__pb2.SubscribeUpdate.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SubscribeReplayInfo(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/geyser.Geyser/SubscribeReplayInfo',
            geyser__pb2.SubscribeReplayInfoRequest.SerializeToString,
            geyser__pb2.SubscribeReplayInfoResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Ping(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/geyser.Geyser/Ping',
            geyser__pb2.PingRequest.SerializeToString,
            geyser__pb2.PongResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetLatestBlockhash(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/geyser.Geyser/GetLatestBlockhash',
            geyser__pb2.GetLatestBlockhashRequest.SerializeToString,
            geyser__pb2.GetLatestBlockhashResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetBlockHeight(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/geyser.Geyser/GetBlockHeight',
            geyser__pb2.GetBlockHeightRequest.SerializeToString,
            geyser__pb2.GetBlockHeightResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetSlot(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/geyser.Geyser/GetSlot',
            geyser__pb2.GetSlotRequest.SerializeToString,
            geyser__pb2.GetSlotResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def IsBlockhashValid(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/geyser.Geyser/IsBlockhashValid',
            geyser__pb2.IsBlockhashValidRequest.SerializeToString,
            geyser__pb2.IsBlockhashValidResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetVersion(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/geyser.Geyser/GetVersion',
            geyser__pb2.GetVersionRequest.SerializeToString,
            geyser__pb2.GetVersionResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



import yellowstone_fumarole_proto.geyser_pb2 as _geyser_pb2
import yellowstone_fumarole_proto.solana_storage_pb2 as _solana_storage_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from collections.abc import Iterable as _Iterable, Mapping as _Mapping
from typing import ClassVar as _ClassVar, Optional as _Optional, Union as _Union
from yellowstone_fumarole_proto.geyser_pb2 import SubscribeRequest as SubscribeRequest
from yellowstone_fumarole_proto.geyser_pb2 import SubscribeRequestFilterAccounts as SubscribeRequestFilterAccounts
from yellowstone_fumarole_proto.geyser_pb2 import SubscribeRequestFilterAccountsFilter as SubscribeRequestFilterAccountsFilter
from yellowstone_fumarole_proto.geyser_pb2 import SubscribeRequestFilterAccountsFilterMemcmp as SubscribeRequestFilterAccountsFilterMemcmp
from yellowstone_fumarole_proto.geyser_pb2 import SubscribeRequestFilterAccountsFilterLamports as SubscribeRequestFilterAccountsFilterLamports
from yellowstone_fumarole_proto.geyser_pb2 import SubscribeRequestFilterSlots as SubscribeRequestFilterSlots
from yellowstone_fumarole_proto.geyser_pb2 import SubscribeRequestFilterTransactions as SubscribeRequestFilterTransactions
from yellowstone_fumarole_proto.geyser_pb2 import SubscribeRequestFilterBlocks as SubscribeRequestFilterBlocks
from yellowstone_fumarole_proto.geyser_pb2 import SubscribeRequestFilterBlocksMeta as SubscribeRequestFilterBlocksMeta
from yellowstone_fumarole_proto.geyser_pb2 import SubscribeRequestFilterEntry as SubscribeRequestFilterEntry
from yellowstone_fumarole_proto.geyser_pb2 import SubscribeRequestAccountsDataSlice as SubscribeRequestAccountsDataSlice
from yellowstone_fumarole_proto.geyser_pb2 import SubscribeRequestPing as SubscribeRequestPing
from yellowstone_fumarole_proto.geyser_pb2 import SubscribeUpdate as SubscribeUpdate
from yellowstone_fumarole_proto.geyser_pb2 import SubscribeUpdateAccount as SubscribeUpdateAccount
from yellowstone_fumarole_proto.geyser_pb2 import SubscribeUpdateAccountInfo as SubscribeUpdateAccountInfo
from yellowstone_fumarole_proto.geyser_pb2 import SubscribeUpdateSlot as SubscribeUpdateSlot
from yellowstone_fumarole_proto.geyser_pb2 import SubscribeUpdateTransaction as SubscribeUpdateTransaction
from yellowstone_fumarole_proto.geyser_pb2 import SubscribeUpdateTransactionInfo as SubscribeUpdateTransactionInfo
from yellowstone_fumarole_proto.geyser_pb2 import SubscribeUpdateTransactionStatus as SubscribeUpdateTransactionStatus
from yellowstone_fumarole_proto.geyser_pb2 import SubscribeUpdateBlock as SubscribeUpdateBlock
from yellowstone_fumarole_proto.geyser_pb2 import SubscribeUpdateBlockMeta as SubscribeUpdateBlockMeta
from yellowstone_fumarole_proto.geyser_pb2 import SubscribeUpdateEntry as SubscribeUpdateEntry
from yellowstone_fumarole_proto.geyser_pb2 import SubscribeUpdatePing as SubscribeUpdatePing
from yellowstone_fumarole_proto.geyser_pb2 import SubscribeUpdatePong as SubscribeUpdatePong
from yellowstone_fumarole_proto.geyser_pb2 import SubscribeReplayInfoRequest as SubscribeReplayInfoRequest
from yellowstone_fumarole_proto.geyser_pb2 import SubscribeReplayInfoResponse as SubscribeReplayInfoResponse
from yellowstone_fumarole_proto.geyser_pb2 import PingRequest as PingRequest
from yellowstone_fumarole_proto.geyser_pb2 import PongResponse as PongResponse
from yellowstone_fumarole_proto.geyser_pb2 import GetLatestBlockhashRequest as GetLatestBlockhashRequest
from yellowstone_fumarole_proto.geyser_pb2 import GetLatestBlockhashResponse as GetLatestBlockhashResponse
from yellowstone_fumarole_proto.geyser_pb2 import GetBlockHeightRequest as GetBlockHeightRequest
from yellowstone_fumarole_proto.geyser_pb2 import GetBlockHeightResponse as GetBlockHeightResponse
from yellowstone_fumarole_proto.geyser_pb2 import GetSlotRequest as GetSlotRequest
from yellowstone_fumarole_proto.geyser_pb2 import GetSlotResponse as GetSlotResponse
from yellowstone_fumarole_proto.geyser_pb2 import GetVersionRequest as GetVersionRequest
from yellowstone_fumarole_proto.geyser_pb2 import GetVersionResponse as GetVersionResponse
from yellowstone_fumarole_proto.geyser_pb2 import IsBlockhashValidRequest as IsBlockhashValidRequest
from yellowstone_fumarole_proto.geyser_pb2 import IsBlockhashValidResponse as IsBlockhashValidResponse
from yellowstone_fumarole_proto.geyser_pb2 import CommitmentLevel as CommitmentLevel
from yellowstone_fumarole_proto.geyser_pb2 import SlotStatus as SlotStatus

DESCRIPTOR: _descriptor.FileDescriptor
PROCESSED: _geyser_pb2.CommitmentLevel
CONFIRMED: _geyser_pb2.CommitmentLevel
FINALIZED: _geyser_pb2.CommitmentLevel
SLOT_PROCESSED: _geyser_pb2.SlotStatus
SLOT_CONFIRMED: _geyser_pb2.SlotStatus
SLOT_FINALIZED: _geyser_pb2.SlotStatus
SLOT_FIRST_SHRED_RECEIVED: _geyser_pb2.SlotStatus
SLOT_COMPLETED: _geyser_pb2.SlotStatus
SLOT_CREATED_BANK: _geyser_pb2.SlotStatus
SLOT_DEAD: _geyser_pb2.SlotStatus

class InitialOffsetPolicy(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    LATEST: _ClassVar[InitialOffsetPolicy]
    FROM_SLOT: _ClassVar[InitialOffsetPolicy]
LATEST: InitialOffsetPolicy
FROM_SLOT: InitialOffsetPolicy

class GetSlotRangeRequest(_message.Message):
    __slots__ = ("blockchain_id",)
    BLOCKCHAIN_ID_FIELD_NUMBER: _ClassVar[int]
    blockchain_id: bytes
    def __init__(self, blockchain_id: _Optional[bytes] = ...) -> None: ...

class GetSlotRangeResponse(_message.Message):
    __slots__ = ("blockchain_id", "min_slot", "max_slot")
    BLOCKCHAIN_ID_FIELD_NUMBER: _ClassVar[int]
    MIN_SLOT_FIELD_NUMBER: _ClassVar[int]
    MAX_SLOT_FIELD_NUMBER: _ClassVar[int]
    blockchain_id: bytes
    min_slot: int
    max_slot: int
    def __init__(self, blockchain_id: _Optional[bytes] = ..., min_slot: _Optional[int] = ..., max_slot: _Optional[int] = ...) -> None: ...

class GetChainTipRequest(_message.Message):
    __slots__ = ("blockchain_id",)
    BLOCKCHAIN_ID_FIELD_NUMBER: _ClassVar[int]
    blockchain_id: bytes
    def __init__(self, blockchain_id: _Optional[bytes] = ...) -> None: ...

class GetChainTipResponse(_message.Message):
    __slots__ = ("blockchain_id", "shard_to_max_offset_map")
    class ShardToMaxOffsetMapEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: int
        value: int
        def __init__(self, key: _Optional[int] = ..., value: _Optional[int] = ...) -> None: ...
    BLOCKCHAIN_ID_FIELD_NUMBER: _ClassVar[int]
    SHARD_TO_MAX_OFFSET_MAP_FIELD_NUMBER: _ClassVar[int]
    blockchain_id: bytes
    shard_to_max_offset_map: _containers.ScalarMap[int, int]
    def __init__(self, blockchain_id: _Optional[bytes] = ..., shard_to_max_offset_map: _Optional[_Mapping[int, int]] = ...) -> None: ...

class VersionRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class VersionResponse(_message.Message):
    __slots__ = ("version",)
    VERSION_FIELD_NUMBER: _ClassVar[int]
    version: str
    def __init__(self, version: _Optional[str] = ...) -> None: ...

class GetConsumerGroupInfoRequest(_message.Message):
    __slots__ = ("consumer_group_name",)
    CONSUMER_GROUP_NAME_FIELD_NUMBER: _ClassVar[int]
    consumer_group_name: str
    def __init__(self, consumer_group_name: _Optional[str] = ...) -> None: ...

class DeleteConsumerGroupRequest(_message.Message):
    __slots__ = ("consumer_group_name",)
    CONSUMER_GROUP_NAME_FIELD_NUMBER: _ClassVar[int]
    consumer_group_name: str
    def __init__(self, consumer_group_name: _Optional[str] = ...) -> None: ...

class DeleteConsumerGroupResponse(_message.Message):
    __slots__ = ("success",)
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    success: bool
    def __init__(self, success: bool = ...) -> None: ...

class ListConsumerGroupsRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class ListConsumerGroupsResponse(_message.Message):
    __slots__ = ("consumer_groups",)
    CONSUMER_GROUPS_FIELD_NUMBER: _ClassVar[int]
    consumer_groups: _containers.RepeatedCompositeFieldContainer[ConsumerGroupInfo]
    def __init__(self, consumer_groups: _Optional[_Iterable[_Union[ConsumerGroupInfo, _Mapping]]] = ...) -> None: ...

class ConsumerGroupInfo(_message.Message):
    __slots__ = ("id", "consumer_group_name", "is_stale", "blockchain_id")
    ID_FIELD_NUMBER: _ClassVar[int]
    CONSUMER_GROUP_NAME_FIELD_NUMBER: _ClassVar[int]
    IS_STALE_FIELD_NUMBER: _ClassVar[int]
    BLOCKCHAIN_ID_FIELD_NUMBER: _ClassVar[int]
    id: str
    consumer_group_name: str
    is_stale: bool
    blockchain_id: bytes
    def __init__(self, id: _Optional[str] = ..., consumer_group_name: _Optional[str] = ..., is_stale: bool = ..., blockchain_id: _Optional[bytes] = ...) -> None: ...

class GetSlotLagInfoRequest(_message.Message):
    __slots__ = ("consumer_group_name",)
    CONSUMER_GROUP_NAME_FIELD_NUMBER: _ClassVar[int]
    consumer_group_name: str
    def __init__(self, consumer_group_name: _Optional[str] = ...) -> None: ...

class BlockFilters(_message.Message):
    __slots__ = ("accounts", "transactions", "entries", "blocks_meta")
    class AccountsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: _geyser_pb2.SubscribeRequestFilterAccounts
        def __init__(self, key: _Optional[str] = ..., value: _Optional[_Union[_geyser_pb2.SubscribeRequestFilterAccounts, _Mapping]] = ...) -> None: ...
    class TransactionsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: _geyser_pb2.SubscribeRequestFilterTransactions
        def __init__(self, key: _Optional[str] = ..., value: _Optional[_Union[_geyser_pb2.SubscribeRequestFilterTransactions, _Mapping]] = ...) -> None: ...
    class EntriesEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: _geyser_pb2.SubscribeRequestFilterEntry
        def __init__(self, key: _Optional[str] = ..., value: _Optional[_Union[_geyser_pb2.SubscribeRequestFilterEntry, _Mapping]] = ...) -> None: ...
    class BlocksMetaEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: _geyser_pb2.SubscribeRequestFilterBlocksMeta
        def __init__(self, key: _Optional[str] = ..., value: _Optional[_Union[_geyser_pb2.SubscribeRequestFilterBlocksMeta, _Mapping]] = ...) -> None: ...
    ACCOUNTS_FIELD_NUMBER: _ClassVar[int]
    TRANSACTIONS_FIELD_NUMBER: _ClassVar[int]
    ENTRIES_FIELD_NUMBER: _ClassVar[int]
    BLOCKS_META_FIELD_NUMBER: _ClassVar[int]
    accounts: _containers.MessageMap[str, _geyser_pb2.SubscribeRequestFilterAccounts]
    transactions: _containers.MessageMap[str, _geyser_pb2.SubscribeRequestFilterTransactions]
    entries: _containers.MessageMap[str, _geyser_pb2.SubscribeRequestFilterEntry]
    blocks_meta: _containers.MessageMap[str, _geyser_pb2.SubscribeRequestFilterBlocksMeta]
    def __init__(self, accounts: _Optional[_Mapping[str, _geyser_pb2.SubscribeRequestFilterAccounts]] = ..., transactions: _Optional[_Mapping[str, _geyser_pb2.SubscribeRequestFilterTransactions]] = ..., entries: _Optional[_Mapping[str, _geyser_pb2.SubscribeRequestFilterEntry]] = ..., blocks_meta: _Optional[_Mapping[str, _geyser_pb2.SubscribeRequestFilterBlocksMeta]] = ...) -> None: ...

class DownloadBlockShard(_message.Message):
    __slots__ = ("blockchain_id", "block_uid", "shard_idx", "blockFilters")
    BLOCKCHAIN_ID_FIELD_NUMBER: _ClassVar[int]
    BLOCK_UID_FIELD_NUMBER: _ClassVar[int]
    SHARD_IDX_FIELD_NUMBER: _ClassVar[int]
    BLOCKFILTERS_FIELD_NUMBER: _ClassVar[int]
    blockchain_id: bytes
    block_uid: bytes
    shard_idx: int
    blockFilters: BlockFilters
    def __init__(self, blockchain_id: _Optional[bytes] = ..., block_uid: _Optional[bytes] = ..., shard_idx: _Optional[int] = ..., blockFilters: _Optional[_Union[BlockFilters, _Mapping]] = ...) -> None: ...

class Ping(_message.Message):
    __slots__ = ("ping_id",)
    PING_ID_FIELD_NUMBER: _ClassVar[int]
    ping_id: int
    def __init__(self, ping_id: _Optional[int] = ...) -> None: ...

class Pong(_message.Message):
    __slots__ = ("ping_id",)
    PING_ID_FIELD_NUMBER: _ClassVar[int]
    ping_id: int
    def __init__(self, ping_id: _Optional[int] = ...) -> None: ...

class DataCommand(_message.Message):
    __slots__ = ("download_block_shard", "filter_update")
    DOWNLOAD_BLOCK_SHARD_FIELD_NUMBER: _ClassVar[int]
    FILTER_UPDATE_FIELD_NUMBER: _ClassVar[int]
    download_block_shard: DownloadBlockShard
    filter_update: BlockFilters
    def __init__(self, download_block_shard: _Optional[_Union[DownloadBlockShard, _Mapping]] = ..., filter_update: _Optional[_Union[BlockFilters, _Mapping]] = ...) -> None: ...

class BlockShardDownloadFinish(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class BlockNotFound(_message.Message):
    __slots__ = ("blockchain_id", "block_uid", "shard_idx")
    BLOCKCHAIN_ID_FIELD_NUMBER: _ClassVar[int]
    BLOCK_UID_FIELD_NUMBER: _ClassVar[int]
    SHARD_IDX_FIELD_NUMBER: _ClassVar[int]
    blockchain_id: bytes
    block_uid: bytes
    shard_idx: int
    def __init__(self, blockchain_id: _Optional[bytes] = ..., block_uid: _Optional[bytes] = ..., shard_idx: _Optional[int] = ...) -> None: ...

class DataError(_message.Message):
    __slots__ = ("not_found",)
    NOT_FOUND_FIELD_NUMBER: _ClassVar[int]
    not_found: BlockNotFound
    def __init__(self, not_found: _Optional[_Union[BlockNotFound, _Mapping]] = ...) -> None: ...

class DataResponse(_message.Message):
    __slots__ = ("update", "block_shard_download_finish")
    UPDATE_FIELD_NUMBER: _ClassVar[int]
    BLOCK_SHARD_DOWNLOAD_FINISH_FIELD_NUMBER: _ClassVar[int]
    update: _geyser_pb2.SubscribeUpdate
    block_shard_download_finish: BlockShardDownloadFinish
    def __init__(self, update: _Optional[_Union[_geyser_pb2.SubscribeUpdate, _Mapping]] = ..., block_shard_download_finish: _Optional[_Union[BlockShardDownloadFinish, _Mapping]] = ...) -> None: ...

class CommitOffset(_message.Message):
    __slots__ = ("offset", "shard_id")
    OFFSET_FIELD_NUMBER: _ClassVar[int]
    SHARD_ID_FIELD_NUMBER: _ClassVar[int]
    offset: int
    shard_id: int
    def __init__(self, offset: _Optional[int] = ..., shard_id: _Optional[int] = ...) -> None: ...

class PollBlockchainHistory(_message.Message):
    __slots__ = ("shard_id", "limit")
    SHARD_ID_FIELD_NUMBER: _ClassVar[int]
    FROM_FIELD_NUMBER: _ClassVar[int]
    LIMIT_FIELD_NUMBER: _ClassVar[int]
    shard_id: int
    limit: int
    def __init__(self, shard_id: _Optional[int] = ..., limit: _Optional[int] = ..., **kwargs) -> None: ...

class BlockchainEvent(_message.Message):
    __slots__ = ("offset", "blockchain_id", "block_uid", "num_shards", "slot", "parent_slot", "commitment_level", "blockchain_shard_id", "dead_error")
    OFFSET_FIELD_NUMBER: _ClassVar[int]
    BLOCKCHAIN_ID_FIELD_NUMBER: _ClassVar[int]
    BLOCK_UID_FIELD_NUMBER: _ClassVar[int]
    NUM_SHARDS_FIELD_NUMBER: _ClassVar[int]
    SLOT_FIELD_NUMBER: _ClassVar[int]
    PARENT_SLOT_FIELD_NUMBER: _ClassVar[int]
    COMMITMENT_LEVEL_FIELD_NUMBER: _ClassVar[int]
    BLOCKCHAIN_SHARD_ID_FIELD_NUMBER: _ClassVar[int]
    DEAD_ERROR_FIELD_NUMBER: _ClassVar[int]
    offset: int
    blockchain_id: bytes
    block_uid: bytes
    num_shards: int
    slot: int
    parent_slot: int
    commitment_level: _geyser_pb2.CommitmentLevel
    blockchain_shard_id: int
    dead_error: str
    def __init__(self, offset: _Optional[int] = ..., blockchain_id: _Optional[bytes] = ..., block_uid: _Optional[bytes] = ..., num_shards: _Optional[int] = ..., slot: _Optional[int] = ..., parent_slot: _Optional[int] = ..., commitment_level: _Optional[_Union[_geyser_pb2.CommitmentLevel, str]] = ..., blockchain_shard_id: _Optional[int] = ..., dead_error: _Optional[str] = ...) -> None: ...

class BlockchainHistory(_message.Message):
    __slots__ = ("events",)
    EVENTS_FIELD_NUMBER: _ClassVar[int]
    events: _containers.RepeatedCompositeFieldContainer[BlockchainEvent]
    def __init__(self, events: _Optional[_Iterable[_Union[BlockchainEvent, _Mapping]]] = ...) -> None: ...

class JoinControlPlane(_message.Message):
    __slots__ = ("consumer_group_name",)
    CONSUMER_GROUP_NAME_FIELD_NUMBER: _ClassVar[int]
    consumer_group_name: str
    def __init__(self, consumer_group_name: _Optional[str] = ...) -> None: ...

class ControlCommand(_message.Message):
    __slots__ = ("initial_join", "commit_offset", "poll_hist", "ping")
    INITIAL_JOIN_FIELD_NUMBER: _ClassVar[int]
    COMMIT_OFFSET_FIELD_NUMBER: _ClassVar[int]
    POLL_HIST_FIELD_NUMBER: _ClassVar[int]
    PING_FIELD_NUMBER: _ClassVar[int]
    initial_join: JoinControlPlane
    commit_offset: CommitOffset
    poll_hist: PollBlockchainHistory
    ping: Ping
    def __init__(self, initial_join: _Optional[_Union[JoinControlPlane, _Mapping]] = ..., commit_offset: _Optional[_Union[CommitOffset, _Mapping]] = ..., poll_hist: _Optional[_Union[PollBlockchainHistory, _Mapping]] = ..., ping: _Optional[_Union[Ping, _Mapping]] = ...) -> None: ...

class ControlResponse(_message.Message):
    __slots__ = ("init", "commit_offset", "poll_hist", "pong")
    INIT_FIELD_NUMBER: _ClassVar[int]
    COMMIT_OFFSET_FIELD_NUMBER: _ClassVar[int]
    POLL_HIST_FIELD_NUMBER: _ClassVar[int]
    PONG_FIELD_NUMBER: _ClassVar[int]
    init: InitialConsumerGroupState
    commit_offset: CommitOffsetResult
    poll_hist: BlockchainHistory
    pong: Pong
    def __init__(self, init: _Optional[_Union[InitialConsumerGroupState, _Mapping]] = ..., commit_offset: _Optional[_Union[CommitOffsetResult, _Mapping]] = ..., poll_hist: _Optional[_Union[BlockchainHistory, _Mapping]] = ..., pong: _Optional[_Union[Pong, _Mapping]] = ...) -> None: ...

class CommitOffsetResult(_message.Message):
    __slots__ = ("offset", "shard_id")
    OFFSET_FIELD_NUMBER: _ClassVar[int]
    SHARD_ID_FIELD_NUMBER: _ClassVar[int]
    offset: int
    shard_id: int
    def __init__(self, offset: _Optional[int] = ..., shard_id: _Optional[int] = ...) -> None: ...

class InitialConsumerGroupState(_message.Message):
    __slots__ = ("blockchain_id", "last_committed_offsets")
    class LastCommittedOffsetsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: int
        value: int
        def __init__(self, key: _Optional[int] = ..., value: _Optional[int] = ...) -> None: ...
    BLOCKCHAIN_ID_FIELD_NUMBER: _ClassVar[int]
    LAST_COMMITTED_OFFSETS_FIELD_NUMBER: _ClassVar[int]
    blockchain_id: bytes
    last_committed_offsets: _containers.ScalarMap[int, int]
    def __init__(self, blockchain_id: _Optional[bytes] = ..., last_committed_offsets: _Optional[_Mapping[int, int]] = ...) -> None: ...

class CreateConsumerGroupResponse(_message.Message):
    __slots__ = ("consumer_group_id",)
    CONSUMER_GROUP_ID_FIELD_NUMBER: _ClassVar[int]
    consumer_group_id: str
    def __init__(self, consumer_group_id: _Optional[str] = ...) -> None: ...

class CreateConsumerGroupRequest(_message.Message):
    __slots__ = ("consumer_group_name", "initial_offset_policy", "from_slot")
    CONSUMER_GROUP_NAME_FIELD_NUMBER: _ClassVar[int]
    INITIAL_OFFSET_POLICY_FIELD_NUMBER: _ClassVar[int]
    FROM_SLOT_FIELD_NUMBER: _ClassVar[int]
    consumer_group_name: str
    initial_offset_policy: InitialOffsetPolicy
    from_slot: int
    def __init__(self, consumer_group_name: _Optional[str] = ..., initial_offset_policy: _Optional[_Union[InitialOffsetPolicy, str]] = ..., from_slot: _Optional[int] = ...) -> None: ...



from collections import OrderedDict


class OrderedSet:
    """
    A simple ordered set implementation that maintains the order of elements.
    """

    def __init__(self, iterable=None):
        self._data = []
        self.inner: OrderedDict = OrderedDict()
        if iterable is not None:
            for item in iterable:
                self.inner[item] = None

    def add(self, item):
        self.inner[item] = None
        # Make sure the item is at the end to maintain order
        self.inner.move_to_end(item, last=True)

    def popfirst(self):
        try:
            return self.inner.popitem(last=False)[0]
        except KeyError as e:
            raise KeyError(f"{OrderedDict.__name__} is empty") from e

    def __contains__(self, item):
        return item in self.inner

    def __iter__(self):
        return iter(self.inner.keys())

    def __len__(self):
        return len(self.inner)

    def __repr__(self):
        return f"OrderedSet({list(self._data)})"



from typing import Any, Mapping


class FumaroleClientError(Exception):
    """Base class for all Fumarole-related exceptions."""
    pass


class SubscribeError(FumaroleClientError):
    """Base class for all errors related to subscription."""
    pass


class DownloadSlotError(SubscribeError):
    """Exception raised for errors in the download slot process."""
    
    def __init__(self, message: str, ctx: Mapping[str, Any]):
        super().__init__(message)
        self.ctx = ctx



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

import yellowstone_fumarole_proto.fumarole_pb2 as fumarole__pb2

GRPC_GENERATED_VERSION = '1.74.0'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in fumarole_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class FumaroleStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetConsumerGroupInfo = channel.unary_unary(
                '/fumarole.Fumarole/GetConsumerGroupInfo',
                request_serializer=fumarole__pb2.GetConsumerGroupInfoRequest.SerializeToString,
                response_deserializer=fumarole__pb2.ConsumerGroupInfo.FromString,
                _registered_method=True)
        self.ListConsumerGroups = channel.unary_unary(
                '/fumarole.Fumarole/ListConsumerGroups',
                request_serializer=fumarole__pb2.ListConsumerGroupsRequest.SerializeToString,
                response_deserializer=fumarole__pb2.ListConsumerGroupsResponse.FromString,
                _registered_method=True)
        self.DeleteConsumerGroup = channel.unary_unary(
                '/fumarole.Fumarole/DeleteConsumerGroup',
                request_serializer=fumarole__pb2.DeleteConsumerGroupRequest.SerializeToString,
                response_deserializer=fumarole__pb2.DeleteConsumerGroupResponse.FromString,
                _registered_method=True)
        self.CreateConsumerGroup = channel.unary_unary(
                '/fumarole.Fumarole/CreateConsumerGroup',
                request_serializer=fumarole__pb2.CreateConsumerGroupRequest.SerializeToString,
                response_deserializer=fumarole__pb2.CreateConsumerGroupResponse.FromString,
                _registered_method=True)
        self.DownloadBlock = channel.unary_stream(
                '/fumarole.Fumarole/DownloadBlock',
                request_serializer=fumarole__pb2.DownloadBlockShard.SerializeToString,
                response_deserializer=fumarole__pb2.DataResponse.FromString,
                _registered_method=True)
        self.SubscribeData = channel.stream_stream(
                '/fumarole.Fumarole/SubscribeData',
                request_serializer=fumarole__pb2.DataCommand.SerializeToString,
                response_deserializer=fumarole__pb2.DataResponse.FromString,
                _registered_method=True)
        self.GetChainTip = channel.unary_unary(
                '/fumarole.Fumarole/GetChainTip',
                request_serializer=fumarole__pb2.GetChainTipRequest.SerializeToString,
                response_deserializer=fumarole__pb2.GetChainTipResponse.FromString,
                _registered_method=True)
        self.Subscribe = channel.stream_stream(
                '/fumarole.Fumarole/Subscribe',
                request_serializer=fumarole__pb2.ControlCommand.SerializeToString,
                response_deserializer=fumarole__pb2.ControlResponse.FromString,
                _registered_method=True)
        self.Version = channel.unary_unary(
                '/fumarole.Fumarole/Version',
                request_serializer=fumarole__pb2.VersionRequest.SerializeToString,
                response_deserializer=fumarole__pb2.VersionResponse.FromString,
                _registered_method=True)
        self.GetSlotRange = channel.unary_unary(
                '/fumarole.Fumarole/GetSlotRange',
                request_serializer=fumarole__pb2.GetSlotRangeRequest.SerializeToString,
                response_deserializer=fumarole__pb2.GetSlotRangeResponse.FromString,
                _registered_method=True)


class FumaroleServicer(object):
    """Missing associated documentation comment in .proto file."""

    def GetConsumerGroupInfo(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListConsumerGroups(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DeleteConsumerGroup(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CreateConsumerGroup(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DownloadBlock(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SubscribeData(self, request_iterator, context):
        """Represents subscription to the data plane
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetChainTip(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Subscribe(self, request_iterator, context):
        """Represents subscription to the control plane
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Version(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetSlotRange(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_FumaroleServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'GetConsumerGroupInfo': grpc.unary_unary_rpc_method_handler(
                    servicer.GetConsumerGroupInfo,
                    request_deserializer=fumarole__pb2.GetConsumerGroupInfoRequest.FromString,
                    response_serializer=fumarole__pb2.ConsumerGroupInfo.SerializeToString,
            ),
            'ListConsumerGroups': grpc.unary_unary_rpc_method_handler(
                    servicer.ListConsumerGroups,
                    request_deserializer=fumarole__pb2.ListConsumerGroupsRequest.FromString,
                    response_serializer=fumarole__pb2.ListConsumerGroupsResponse.SerializeToString,
            ),
            'DeleteConsumerGroup': grpc.unary_unary_rpc_method_handler(
                    servicer.DeleteConsumerGroup,
                    request_deserializer=fumarole__pb2.DeleteConsumerGroupRequest.FromString,
                    response_serializer=fumarole__pb2.DeleteConsumerGroupResponse.SerializeToString,
            ),
            'CreateConsumerGroup': grpc.unary_unary_rpc_method_handler(
                    servicer.CreateConsumerGroup,
                    request_deserializer=fumarole__pb2.CreateConsumerGroupRequest.FromString,
                    response_serializer=fumarole__pb2.CreateConsumerGroupResponse.SerializeToString,
            ),
            'DownloadBlock': grpc.unary_stream_rpc_method_handler(
                    servicer.DownloadBlock,
                    request_deserializer=fumarole__pb2.DownloadBlockShard.FromString,
                    response_serializer=fumarole__pb2.DataResponse.SerializeToString,
            ),
            'SubscribeData': grpc.stream_stream_rpc_method_handler(
                    servicer.SubscribeData,
                    request_deserializer=fumarole__pb2.DataCommand.FromString,
                    response_serializer=fumarole__pb2.DataResponse.SerializeToString,
            ),
            'GetChainTip': grpc.unary_unary_rpc_method_handler(
                    servicer.GetChainTip,
                    request_deserializer=fumarole__pb2.GetChainTipRequest.FromString,
                    response_serializer=fumarole__pb2.GetChainTipResponse.SerializeToString,
            ),
            'Subscribe': grpc.stream_stream_rpc_method_handler(
                    servicer.Subscribe,
                    request_deserializer=fumarole__pb2.ControlCommand.FromString,
                    response_serializer=fumarole__pb2.ControlResponse.SerializeToString,
            ),
            'Version': grpc.unary_unary_rpc_method_handler(
                    servicer.Version,
                    request_deserializer=fumarole__pb2.VersionRequest.FromString,
                    response_serializer=fumarole__pb2.VersionResponse.SerializeToString,
            ),
            'GetSlotRange': grpc.unary_unary_rpc_method_handler(
                    servicer.GetSlotRange,
                    request_deserializer=fumarole__pb2.GetSlotRangeRequest.FromString,
                    response_serializer=fumarole__pb2.GetSlotRangeResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'fumarole.Fumarole', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('fumarole.Fumarole', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class Fumarole(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def GetConsumerGroupInfo(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/fumarole.Fumarole/GetConsumerGroupInfo',
            fumarole__pb2.GetConsumerGroupInfoRequest.SerializeToString,
            fumarole__pb2.ConsumerGroupInfo.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ListConsumerGroups(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/fumarole.Fumarole/ListConsumerGroups',
            fumarole__pb2.ListConsumerGroupsRequest.SerializeToString,
            fumarole__pb2.ListConsumerGroupsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def DeleteConsumerGroup(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/fumarole.Fumarole/DeleteConsumerGroup',
            fumarole__pb2.DeleteConsumerGroupRequest.SerializeToString,
            fumarole__pb2.DeleteConsumerGroupResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def CreateConsumerGroup(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/fumarole.Fumarole/CreateConsumerGroup',
            fumarole__pb2.CreateConsumerGroupRequest.SerializeToString,
            fumarole__pb2.CreateConsumerGroupResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def DownloadBlock(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/fumarole.Fumarole/DownloadBlock',
            fumarole__pb2.DownloadBlockShard.SerializeToString,
            fumarole__pb2.DataResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SubscribeData(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(
            request_iterator,
            target,
            '/fumarole.Fumarole/SubscribeData',
            fumarole__pb2.DataCommand.SerializeToString,
            fumarole__pb2.DataResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetChainTip(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/fumarole.Fumarole/GetChainTip',
            fumarole__pb2.GetChainTipRequest.SerializeToString,
            fumarole__pb2.GetChainTipResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Subscribe(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(
            request_iterator,
            target,
            '/fumarole.Fumarole/Subscribe',
            fumarole__pb2.ControlCommand.SerializeToString,
            fumarole__pb2.ControlResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Version(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/fumarole.Fumarole/Version',
            fumarole__pb2.VersionRequest.SerializeToString,
            fumarole__pb2.VersionResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetSlotRange(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/fumarole.Fumarole/GetSlotRange',
            fumarole__pb2.GetSlotRangeRequest.SerializeToString,
            fumarole__pb2.GetSlotRangeResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



# DataPlaneConn
from abc import abstractmethod, ABC
import asyncio
from contextlib import suppress
from typing import Any

import grpc
from typing import Literal, Mapping, Optional
from collections import deque
from dataclasses import dataclass
import time
from yellowstone_fumarole_client.runtime.state_machine import (
    FumaroleSM,
    FumeDownloadRequest,
    FumeOffset,
    FumeShardIdx,
)
from yellowstone_fumarole_proto.geyser_pb2 import (
    SubscribeRequest,
    SubscribeUpdate,
    SubscribeUpdateSlot,
)
from yellowstone_fumarole_proto.fumarole_pb2 import (
    ControlCommand,
    PollBlockchainHistory,
    CommitOffset,
    ControlResponse,
    DownloadBlockShard,
    BlockFilters,
)
from yellowstone_fumarole_proto.fumarole_pb2_grpc import (
    FumaroleStub,
)
from yellowstone_fumarole_client.utils.aio import Interval
from yellowstone_fumarole_client.error import SubscribeError, DownloadSlotError
import logging


# Constants
DEFAULT_GC_INTERVAL = 5

DEFAULT_SLOT_MEMORY_RETENTION = 10000


# DownloadTaskResult
@dataclass
class CompletedDownloadBlockTask:
    """Represents a completed download block task."""

    slot: int
    block_uid: bytes
    shard_idx: FumeShardIdx
    total_event_downloaded: int


DownloadBlockErrorKind = Literal[
    "Disconnected",
    "OutletDisconnected",
    "BlockShardNotFound",
    "FailedDownload",
    "Fatal",
    "SessionDeprecated"
]

@dataclass
class DownloadBlockError:
    """Represents an error that occurred during the download of a block."""
    kind: DownloadBlockErrorKind
    message: str
    ctx: Mapping[str, Any]

    def into_subscribe_error(self) -> SubscribeError:
        """Convert the error into a SubscribeError."""
        return DownloadSlotError(
            f"{self.kind}: {self.message}",
            ctx=self.ctx
        )

@dataclass
class DownloadTaskResult:
    """Represents the result of a download task."""

    kind: str  # 'Ok' or 'Err'
    completed: Optional[CompletedDownloadBlockTask] = None
    slot: Optional[int] = None
    err: Optional[DownloadBlockError] = None


LOGGER = logging.getLogger(__name__)


class AsyncSlotDownloader(ABC):
    """Abstract base class for slot downloaders."""

    @abstractmethod
    async def run_download(
        self, subscribe_request: SubscribeRequest, spec: "DownloadTaskArgs"
    ) -> DownloadTaskResult:
        """Run the download task for a given slot."""
        pass


SUBSCRIBE_REQ_UPDATE_TYPE_MARKER: int = 1
CONTROL_PLANE_RESP_TYPE_MARKER: int = 2
COMMIT_TICK_TYPE_MARKER: int = 3
DOWNLOAD_TASK_TYPE_MARKER: int = 4


# TokioFumeDragonsmouthRuntime
class AsyncioFumeDragonsmouthRuntime:
    """Asynchronous runtime for Fumarole with Dragonsmouth-like stream support."""

    def __init__(
        self,
        sm: FumaroleSM,
        slot_downloader: AsyncSlotDownloader,
        subscribe_request_update_q: asyncio.Queue,
        subscribe_request: SubscribeRequest,
        consumer_group_name: str,
        control_plane_tx_q: asyncio.Queue,
        control_plane_rx_q: asyncio.Queue,
        dragonsmouth_outlet: asyncio.Queue,
        commit_interval: float,  # in seconds
        gc_interval: int,
        max_concurrent_download: int = 10,
    ):
        """Initialize the runtime with the given parameters.

        Args:
            sm (FumaroleSM): The state machine managing the Fumarole state.
            slot_downloader (AsyncSlotDownloader): The downloader for slots.
            subscribe_request_update_q (asyncio.Queue): The queue for subscribe request updates.
            subscribe_request (SubscribeRequest): The initial subscribe request.
            consumer_group_name (str): The name of the consumer group.
            control_plane_tx_q (asyncio.Queue): The queue for sending control commands.
            control_plane_rx_q (asyncio.Queue): The queue for receiving control responses.
            dragonsmouth_outlet (asyncio.Queue): The outlet for Dragonsmouth updates.
            commit_interval (float): The interval for committing offsets, in seconds.
            gc_interval (int): The interval for garbage collection, in seconds.
            max_concurrent_download (int): The maximum number of concurrent download tasks.
        """
        self.sm = sm
        self.slot_downloader: AsyncSlotDownloader = slot_downloader
        self.subscribe_request_update_rx: asyncio.Queue = subscribe_request_update_q
        self.subscribe_request = subscribe_request
        self.consumer_group_name = consumer_group_name
        self.control_plane_tx: asyncio.Queue = control_plane_tx_q
        self.control_plane_rx: asyncio.Queue = control_plane_rx_q
        self.dragonsmouth_outlet: asyncio.Queue = dragonsmouth_outlet
        self.commit_interval = commit_interval
        self.gc_interval = gc_interval
        self.max_concurrent_download = max_concurrent_download
        self.poll_hist_inflight = False
        self.commit_offset_inflight = False

        # holds metadata about the download task
        self.download_tasks = dict()
        self.inflight_tasks = dict()
        self.successful_download_cnt = 0
        self.is_closed = False

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.aclose()

    async def aclose(self):
        self.control_plane_tx.shutdown()
        self.dragonsmouth_outlet.shutdown()
        for t, kind in self.inflight_tasks.items():
            LOGGER.debug(f"closing {kind} task")
            t.cancel()

    def _build_poll_history_cmd(
        self, from_offset: Optional[FumeOffset]
    ) -> ControlCommand:
        """Build a command to poll the blockchain history."""
        return ControlCommand(poll_hist=PollBlockchainHistory(shard_id=0, limit=None))

    def _build_commit_offset_cmd(self, offset: FumeOffset) -> ControlCommand:
        return ControlCommand(commit_offset=CommitOffset(offset=offset, shard_id=0))

    def _handle_control_response(self, control_response: ControlResponse):
        """Handle the control response received from the control plane."""
        response_field = control_response.WhichOneof("response")
        assert response_field is not None, "Control response is empty"

        match response_field:
            case "poll_hist":
                self.poll_hist_inflight = False
                poll_hist = control_response.poll_hist
                LOGGER.debug(f"Received poll history {len(poll_hist.events)} events")
                self.sm.queue_blockchain_event(poll_hist.events)
            case "commit_offset":
                self.commit_offset_inflight = False
                commit_offset = control_response.commit_offset
                LOGGER.debug(f"Received commit offset: {commit_offset}")
                self.sm.update_committed_offset(commit_offset.offset)
            case "pong":
                LOGGER.debug("Received pong")
            case _:
                raise ValueError("Unexpected control response")

    async def poll_history_if_needed(self):
        """Poll the history if the state machine needs new events."""
        if self.poll_hist_inflight:
            return
        if self.sm.need_new_blockchain_events():
            cmd = self._build_poll_history_cmd(self.sm.committable_offset)
            await self.control_plane_tx.put(cmd)
            self.poll_hist_inflight = True

    def commitment_level(self):
        """Gets the commitment level from the subscribe request."""
        return self.subscribe_request.commitment

    def _schedule_download_task_if_any(self):
        """Schedules download tasks if there are any available slots."""
        while True:
            LOGGER.debug("Checking for download tasks to schedule")
            if len(self.download_tasks) >= self.max_concurrent_download:
                break

            # Pop a slot to download from the state machine
            LOGGER.debug("Popping slot to download")
            download_request = self.sm.pop_slot_to_download(self.commitment_level())
            if not download_request:
                LOGGER.debug("No download request available")
                break

            LOGGER.debug(f"Download request for slot {download_request.slot} popped")
            assert (
                download_request.blockchain_id
            ), "Download request must have a blockchain ID"
            download_task_args = DownloadTaskArgs(
                download_request=download_request,
                dragonsmouth_outlet=self.dragonsmouth_outlet,
            )

            coro = self.slot_downloader.run_download(
                self.subscribe_request, download_task_args
            )
            download_task = asyncio.create_task(coro)
            self.download_tasks[download_task] = download_request
            self.inflight_tasks[download_task] = DOWNLOAD_TASK_TYPE_MARKER
            LOGGER.debug(f"Scheduling download task for slot {download_request.slot}")

    async def _handle_download_result(self, download_result: DownloadTaskResult):
        """Handles the result of a download task."""
        if download_result.kind == "Ok":
            self.successful_download_cnt += 1
            completed = download_result.completed
            LOGGER.debug(
                f"Download({self.successful_download_cnt}) completed for slot {completed.slot}, shard {completed.shard_idx}, {completed.total_event_downloaded} total events"
            )
            self.sm.make_slot_download_progress(completed.slot, completed.shard_idx)
        else:
            slot = download_result.slot
            err_kind = download_result.err.kind
            LOGGER.error(f"Download error for slot {slot}: {download_result.err}")
            # If the client queue is disconnected, we don't do anything, next run iteration will close anyway.
            self.is_closed = True
            if err_kind != "OutletDisconnected":
                with suppress(asyncio.QueueShutDown):
                    await self.dragonsmouth_outlet.put(download_result.err.into_subscribe_error())

    async def _force_commit_offset(self):
        LOGGER.debug(f"Force committing offset {self.sm.committable_offset}")
        await self.control_plane_tx.put(
            self._build_commit_offset_cmd(self.sm.committable_offset)
        )

    async def _commit_offset(self):
        self.last_commit = time.time()
        if self.commit_offset_inflight:
            return
        if self.sm.last_committed_offset < self.sm.committable_offset:
            LOGGER.debug(f"Committing offset {self.sm.committable_offset}")
            await self._force_commit_offset()
            self.commit_offset_inflight = True

    async def _drain_slot_status(self):
        """Drains the slot status from the state machine and sends updates to the Dragonsmouth outlet."""
        commitment = self.subscribe_request.commitment
        slot_status_vec = deque()
        while slot_status := self.sm.pop_next_slot_status():
            slot_status_vec.append(slot_status)

        if not slot_status_vec:
            return

        LOGGER.debug(f"Draining {len(slot_status_vec)} slot status")
        for slot_status in slot_status_vec:
            self.sm.mark_event_as_processed(slot_status.session_sequence)
            matched_filters = []
            for filter_name, filter in self.subscribe_request.slots.items():
                if (
                    filter.filter_by_commitment
                    and slot_status.commitment_level == commitment
                ):
                    matched_filters.append(filter_name)
                elif not filter.filter_by_commitment:
                    matched_filters.append(filter_name)
            if matched_filters:
                LOGGER.debug(
                    f"Matched {len(matched_filters)} filters for SlotStatus Update"
                )
                update = SubscribeUpdate(
                    filters=matched_filters,
                    created_at=None,
                    slot=SubscribeUpdateSlot(
                        slot=slot_status.slot,
                        parent=slot_status.parent_slot,
                        status=slot_status.commitment_level,
                        dead_error=slot_status.dead_error,
                    ),
                )
                await self.dragonsmouth_outlet.put(update)


    async def _handle_control_plane_resp(
        self, result: ControlResponse | Exception
    ) -> bool:
        """Handles the control plane response."""
        if isinstance(result, Exception):
            await self.dragonsmouth_outlet.put(result)
            return False
        self._handle_control_response(result)
        return True

    def handle_new_subscribe_request(self, subscribe_request: SubscribeRequest):
        self.subscribe_request = subscribe_request

    async def run(self):
        """Runs the Fumarole asyncio runtime."""
        LOGGER.debug(f"Fumarole runtime starting...")
        await self.control_plane_tx.put(self._build_poll_history_cmd(None))
        LOGGER.debug("Initial poll history command sent")
        await self._force_commit_offset()
        LOGGER.debug("Initial commit offset command sent")
        ticks = 0

        self.inflight_tasks = {
            asyncio.create_task(
                self.subscribe_request_update_rx.get()
            ): SUBSCRIBE_REQ_UPDATE_TYPE_MARKER,
            asyncio.create_task(
                self.control_plane_rx.get()
            ): CONTROL_PLANE_RESP_TYPE_MARKER,
            asyncio.create_task(
                Interval(self.commit_interval).tick()
            ): COMMIT_TICK_TYPE_MARKER,
        }

        while self.inflight_tasks and not self.is_closed:
            ticks += 1
            LOGGER.debug(f"Runtime loop tick")
            if ticks % self.gc_interval == 0:
                LOGGER.debug("Running garbage collection")
                self.sm.gc()
                ticks = 0
            LOGGER.debug(f"Polling history if needed")
            await self.poll_history_if_needed()
            LOGGER.debug("Scheduling download tasks if any")
            self._schedule_download_task_if_any()
            download_task_inflight = len(self.download_tasks)
            LOGGER.debug(
                f"Current download tasks in flight: {download_task_inflight} / {self.max_concurrent_download}"
            )
            done, _pending = await asyncio.wait(
                self.inflight_tasks.keys(), return_when=asyncio.FIRST_COMPLETED
            )
            for t in done:
                result = t.result()
                sigcode = self.inflight_tasks.pop(t)
                if sigcode == SUBSCRIBE_REQ_UPDATE_TYPE_MARKER:
                    LOGGER.debug("Dragonsmouth subscribe request received")
                    assert isinstance(
                        result, SubscribeRequest
                    ), "Expected SubscribeRequest"
                    self.handle_new_subscribe_request(result)
                    new_task = asyncio.create_task(
                        self.subscribe_request_update_rx.get()
                    )
                    self.inflight_tasks[new_task] = SUBSCRIBE_REQ_UPDATE_TYPE_MARKER
                    pass
                elif sigcode == CONTROL_PLANE_RESP_TYPE_MARKER:
                    LOGGER.debug("Control plane response received")
                    if not await self._handle_control_plane_resp(result):
                        LOGGER.debug("Control plane error")
                        return
                    new_task = asyncio.create_task(self.control_plane_rx.get())
                    self.inflight_tasks[new_task] = CONTROL_PLANE_RESP_TYPE_MARKER
                elif sigcode == DOWNLOAD_TASK_TYPE_MARKER:
                    LOGGER.debug("Download task result received")
                    assert self.download_tasks.pop(t)
                    await self._handle_download_result(result)
                elif sigcode == COMMIT_TICK_TYPE_MARKER:
                    LOGGER.debug("Commit tick reached")
                    await self._commit_offset()
                    new_task = asyncio.create_task(
                        Interval(self.commit_interval).tick()
                    )
                    self.inflight_tasks[new_task] = COMMIT_TICK_TYPE_MARKER
                else:
                    raise RuntimeError(f"Unexpected task name: {sigcode}")

            await self._drain_slot_status()
        await self.aclose()

# DownloadTaskRunnerChannels
@dataclass
class DownloadTaskRunnerChannels:
    download_task_queue_tx: asyncio.Queue
    cnc_tx: asyncio.Queue
    download_result_rx: asyncio.Queue


# DownloadTaskRunnerCommand
@dataclass
class DownloadTaskRunnerCommand:
    kind: str
    subscribe_request: Optional[SubscribeRequest] = None

    @classmethod
    def UpdateSubscribeRequest(cls, subscribe_request: SubscribeRequest):
        return cls(kind="UpdateSubscribeRequest", subscribe_request=subscribe_request)


# DownloadTaskArgs
@dataclass
class DownloadTaskArgs:
    download_request: FumeDownloadRequest
    dragonsmouth_outlet: asyncio.Queue


class GrpcSlotDownloader(AsyncSlotDownloader):

    def __init__(
        self,
        client: FumaroleStub,
    ):
        self.client = client

    async def run_download(
        self, subscribe_request: SubscribeRequest, spec: DownloadTaskArgs
    ) -> DownloadTaskResult:

        download_task = GrpcDownloadBlockTaskRun(
            download_request=spec.download_request,
            client=self.client,
            filters=BlockFilters(
                accounts=subscribe_request.accounts,
                transactions=subscribe_request.transactions,
                entries=subscribe_request.entry,
                blocks_meta=subscribe_request.blocks_meta,
            ),
            dragonsmouth_oulet=spec.dragonsmouth_outlet,
        )

        LOGGER.debug(f"Running download task for slot {spec.download_request.slot}")
        return await download_task.run()


# GrpcDownloadBlockTaskRun
class GrpcDownloadBlockTaskRun:
    def __init__(
        self,
        download_request: FumeDownloadRequest,
        client: FumaroleStub,
        filters: Optional[BlockFilters],
        dragonsmouth_oulet: asyncio.Queue,
    ):
        self.download_request = download_request
        self.client = client
        self.filters = filters
        self.dragonsmouth_oulet = dragonsmouth_oulet

    def map_tonic_error_code_to_download_block_error(
        self,
        e: grpc.aio.AioRpcError
    ) -> DownloadBlockError:
        code = e.code()
        ctx = {
            "grpc_status_code": code,
            "slot": self.download_request.slot,
            "block_uid": self.download_request.block_uid,
        }
        if code == grpc.StatusCode.NOT_FOUND:
            return DownloadBlockError(
                kind="BlockShardNotFound", message="Block shard not found", ctx=ctx,
            )
        elif code == grpc.StatusCode.UNAVAILABLE:
            return DownloadBlockError(kind="Disconnected", message="Disconnected", ctx=ctx)
        elif code in (
            grpc.StatusCode.INTERNAL,
            grpc.StatusCode.ABORTED,
            grpc.StatusCode.DATA_LOSS,
            grpc.StatusCode.RESOURCE_EXHAUSTED,
            grpc.StatusCode.UNKNOWN,
            grpc.StatusCode.CANCELLED,
            grpc.StatusCode.DEADLINE_EXCEEDED,
        ):
            return DownloadBlockError(kind="FailedDownload", message="Failed download", ctx=ctx)
        elif code == grpc.StatusCode.INVALID_ARGUMENT:
            raise ValueError("Invalid argument")
        elif code == grpc.StatusCode.FAILED_PRECONDITION:
            return DownloadBlockError(kind="SessionDeprecated", message="Session is deprecated", ctx=ctx)
        else:
            return DownloadBlockError(kind="Fatal", message=f"Unknown error: {code}", ctx=ctx)

    async def run(self) -> DownloadTaskResult:
        request = DownloadBlockShard(
            blockchain_id=self.download_request.blockchain_id,
            block_uid=self.download_request.block_uid,
            shard_idx=0,
            blockFilters=self.filters,
        )
        try:
            LOGGER.debug(
                f"Requesting download for block {self.download_request.block_uid.hex()} at slot {self.download_request.slot}"
            )
            resp = self.client.DownloadBlock(request)
        except grpc.aio.AioRpcError as e:
            LOGGER.error(f"Download block error: {e}")
            return DownloadTaskResult(
                kind="Err",
                slot=self.download_request.slot,
                err=self.map_tonic_error_code_to_download_block_error(e),
            )

        total_event_downloaded = 0
        try:
            async for data in resp:
                kind = data.WhichOneof("response")
                match kind:
                    case "update":
                        update = data.update
                        assert update is not None, "Update is None"
                        total_event_downloaded += 1
                        try:
                            await self.dragonsmouth_oulet.put(update)
                        except asyncio.QueueShutDown:
                            LOGGER.error("Dragonsmouth outlet is disconnected")
                            return DownloadTaskResult(
                                kind="Err",
                                slot=self.download_request.slot,
                                err=DownloadBlockError(
                                    kind="OutletDisconnected",
                                    message="Outlet disconnected",
                                    ctx = {
                                        "slot": self.download_request.slot,
                                        "block_uid": self.download_request.block_uid,
                                    }
                                ),
                            )
                    case "block_shard_download_finish":
                        LOGGER.debug(
                            f"Download finished for block {self.download_request.block_uid.hex()} at slot {self.download_request.slot}"
                        )
                        return DownloadTaskResult(
                            kind="Ok",
                            completed=CompletedDownloadBlockTask(
                                slot=self.download_request.slot,
                                block_uid=self.download_request.block_uid,
                                shard_idx=0,
                                total_event_downloaded=total_event_downloaded,
                            ),
                        )
                    case unknown:
                        raise RuntimeError(f"Unexpected response kind: {unknown}")

        except grpc.aio.AioRpcError as e:
            e2 = self.map_tonic_error_code_to_download_block_error(e)
            LOGGER.error(f"Download block error: {e}, {e2}")
            return DownloadTaskResult(
                kind="Err",
                slot=self.download_request.slot,
                err=e2,
            )

        return DownloadTaskResult(
            kind="Err",
            slot=self.download_request.slot,
            err=DownloadBlockError(
                kind="FailedDownload",
                message="Failed download",
                ctx={
                    "slot": self.download_request.slot,
                    "block_uid": self.download_request.block_uid,
                },
            ),
        )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: geyser.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    1,
    '',
    'geyser.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
import yellowstone_fumarole_proto.solana_storage_pb2 as solana__storage__pb2

from yellowstone_fumarole_proto.solana_storage_pb2 import *

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0cgeyser.proto\x12\x06geyser\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x14solana-storage.proto\"\x9c\n\n\x10SubscribeRequest\x12\x38\n\x08\x61\x63\x63ounts\x18\x01 \x03(\x0b\x32&.geyser.SubscribeRequest.AccountsEntry\x12\x32\n\x05slots\x18\x02 \x03(\x0b\x32#.geyser.SubscribeRequest.SlotsEntry\x12@\n\x0ctransactions\x18\x03 \x03(\x0b\x32*.geyser.SubscribeRequest.TransactionsEntry\x12M\n\x13transactions_status\x18\n \x03(\x0b\x32\x30.geyser.SubscribeRequest.TransactionsStatusEntry\x12\x34\n\x06\x62locks\x18\x04 \x03(\x0b\x32$.geyser.SubscribeRequest.BlocksEntry\x12=\n\x0b\x62locks_meta\x18\x05 \x03(\x0b\x32(.geyser.SubscribeRequest.BlocksMetaEntry\x12\x32\n\x05\x65ntry\x18\x08 \x03(\x0b\x32#.geyser.SubscribeRequest.EntryEntry\x12\x30\n\ncommitment\x18\x06 \x01(\x0e\x32\x17.geyser.CommitmentLevelH\x00\x88\x01\x01\x12\x46\n\x13\x61\x63\x63ounts_data_slice\x18\x07 \x03(\x0b\x32).geyser.SubscribeRequestAccountsDataSlice\x12/\n\x04ping\x18\t \x01(\x0b\x32\x1c.geyser.SubscribeRequestPingH\x01\x88\x01\x01\x12\x16\n\tfrom_slot\x18\x0b \x01(\x04H\x02\x88\x01\x01\x1aW\n\rAccountsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x35\n\x05value\x18\x02 \x01(\x0b\x32&.geyser.SubscribeRequestFilterAccounts:\x02\x38\x01\x1aQ\n\nSlotsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x32\n\x05value\x18\x02 \x01(\x0b\x32#.geyser.SubscribeRequestFilterSlots:\x02\x38\x01\x1a_\n\x11TransactionsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x39\n\x05value\x18\x02 \x01(\x0b\x32*.geyser.SubscribeRequestFilterTransactions:\x02\x38\x01\x1a\x65\n\x17TransactionsStatusEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x39\n\x05value\x18\x02 \x01(\x0b\x32*.geyser.SubscribeRequestFilterTransactions:\x02\x38\x01\x1aS\n\x0b\x42locksEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x33\n\x05value\x18\x02 \x01(\x0b\x32$.geyser.SubscribeRequestFilterBlocks:\x02\x38\x01\x1a[\n\x0f\x42locksMetaEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x37\n\x05value\x18\x02 \x01(\x0b\x32(.geyser.SubscribeRequestFilterBlocksMeta:\x02\x38\x01\x1aQ\n\nEntryEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x32\n\x05value\x18\x02 \x01(\x0b\x32#.geyser.SubscribeRequestFilterEntry:\x02\x38\x01\x42\r\n\x0b_commitmentB\x07\n\x05_pingB\x0c\n\n_from_slot\"\xbf\x01\n\x1eSubscribeRequestFilterAccounts\x12\x0f\n\x07\x61\x63\x63ount\x18\x02 \x03(\t\x12\r\n\x05owner\x18\x03 \x03(\t\x12=\n\x07\x66ilters\x18\x04 \x03(\x0b\x32,.geyser.SubscribeRequestFilterAccountsFilter\x12#\n\x16nonempty_txn_signature\x18\x05 \x01(\x08H\x00\x88\x01\x01\x42\x19\n\x17_nonempty_txn_signature\"\xf3\x01\n$SubscribeRequestFilterAccountsFilter\x12\x44\n\x06memcmp\x18\x01 \x01(\x0b\x32\x32.geyser.SubscribeRequestFilterAccountsFilterMemcmpH\x00\x12\x12\n\x08\x64\x61tasize\x18\x02 \x01(\x04H\x00\x12\x1d\n\x13token_account_state\x18\x03 \x01(\x08H\x00\x12H\n\x08lamports\x18\x04 \x01(\x0b\x32\x34.geyser.SubscribeRequestFilterAccountsFilterLamportsH\x00\x42\x08\n\x06\x66ilter\"y\n*SubscribeRequestFilterAccountsFilterMemcmp\x12\x0e\n\x06offset\x18\x01 \x01(\x04\x12\x0f\n\x05\x62ytes\x18\x02 \x01(\x0cH\x00\x12\x10\n\x06\x62\x61se58\x18\x03 \x01(\tH\x00\x12\x10\n\x06\x62\x61se64\x18\x04 \x01(\tH\x00\x42\x06\n\x04\x64\x61ta\"m\n,SubscribeRequestFilterAccountsFilterLamports\x12\x0c\n\x02\x65q\x18\x01 \x01(\x04H\x00\x12\x0c\n\x02ne\x18\x02 \x01(\x04H\x00\x12\x0c\n\x02lt\x18\x03 \x01(\x04H\x00\x12\x0c\n\x02gt\x18\x04 \x01(\x04H\x00\x42\x05\n\x03\x63mp\"\x8f\x01\n\x1bSubscribeRequestFilterSlots\x12!\n\x14\x66ilter_by_commitment\x18\x01 \x01(\x08H\x00\x88\x01\x01\x12\x1e\n\x11interslot_updates\x18\x02 \x01(\x08H\x01\x88\x01\x01\x42\x17\n\x15_filter_by_commitmentB\x14\n\x12_interslot_updates\"\xd2\x01\n\"SubscribeRequestFilterTransactions\x12\x11\n\x04vote\x18\x01 \x01(\x08H\x00\x88\x01\x01\x12\x13\n\x06\x66\x61iled\x18\x02 \x01(\x08H\x01\x88\x01\x01\x12\x16\n\tsignature\x18\x05 \x01(\tH\x02\x88\x01\x01\x12\x17\n\x0f\x61\x63\x63ount_include\x18\x03 \x03(\t\x12\x17\n\x0f\x61\x63\x63ount_exclude\x18\x04 \x03(\t\x12\x18\n\x10\x61\x63\x63ount_required\x18\x06 \x03(\tB\x07\n\x05_voteB\t\n\x07_failedB\x0c\n\n_signature\"\xd9\x01\n\x1cSubscribeRequestFilterBlocks\x12\x17\n\x0f\x61\x63\x63ount_include\x18\x01 \x03(\t\x12!\n\x14include_transactions\x18\x02 \x01(\x08H\x00\x88\x01\x01\x12\x1d\n\x10include_accounts\x18\x03 \x01(\x08H\x01\x88\x01\x01\x12\x1c\n\x0finclude_entries\x18\x04 \x01(\x08H\x02\x88\x01\x01\x42\x17\n\x15_include_transactionsB\x13\n\x11_include_accountsB\x12\n\x10_include_entries\"\"\n SubscribeRequestFilterBlocksMeta\"\x1d\n\x1bSubscribeRequestFilterEntry\"C\n!SubscribeRequestAccountsDataSlice\x12\x0e\n\x06offset\x18\x01 \x01(\x04\x12\x0e\n\x06length\x18\x02 \x01(\x04\"\"\n\x14SubscribeRequestPing\x12\n\n\x02id\x18\x01 \x01(\x05\"\xb5\x04\n\x0fSubscribeUpdate\x12\x0f\n\x07\x66ilters\x18\x01 \x03(\t\x12\x31\n\x07\x61\x63\x63ount\x18\x02 \x01(\x0b\x32\x1e.geyser.SubscribeUpdateAccountH\x00\x12+\n\x04slot\x18\x03 \x01(\x0b\x32\x1b.geyser.SubscribeUpdateSlotH\x00\x12\x39\n\x0btransaction\x18\x04 \x01(\x0b\x32\".geyser.SubscribeUpdateTransactionH\x00\x12\x46\n\x12transaction_status\x18\n \x01(\x0b\x32(.geyser.SubscribeUpdateTransactionStatusH\x00\x12-\n\x05\x62lock\x18\x05 \x01(\x0b\x32\x1c.geyser.SubscribeUpdateBlockH\x00\x12+\n\x04ping\x18\x06 \x01(\x0b\x32\x1b.geyser.SubscribeUpdatePingH\x00\x12+\n\x04pong\x18\t \x01(\x0b\x32\x1b.geyser.SubscribeUpdatePongH\x00\x12\x36\n\nblock_meta\x18\x07 \x01(\x0b\x32 .geyser.SubscribeUpdateBlockMetaH\x00\x12-\n\x05\x65ntry\x18\x08 \x01(\x0b\x32\x1c.geyser.SubscribeUpdateEntryH\x00\x12.\n\ncreated_at\x18\x0b \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x0e\n\x0cupdate_oneof\"o\n\x16SubscribeUpdateAccount\x12\x33\n\x07\x61\x63\x63ount\x18\x01 \x01(\x0b\x32\".geyser.SubscribeUpdateAccountInfo\x12\x0c\n\x04slot\x18\x02 \x01(\x04\x12\x12\n\nis_startup\x18\x03 \x01(\x08\"\xc8\x01\n\x1aSubscribeUpdateAccountInfo\x12\x0e\n\x06pubkey\x18\x01 \x01(\x0c\x12\x10\n\x08lamports\x18\x02 \x01(\x04\x12\r\n\x05owner\x18\x03 \x01(\x0c\x12\x12\n\nexecutable\x18\x04 \x01(\x08\x12\x12\n\nrent_epoch\x18\x05 \x01(\x04\x12\x0c\n\x04\x64\x61ta\x18\x06 \x01(\x0c\x12\x15\n\rwrite_version\x18\x07 \x01(\x04\x12\x1a\n\rtxn_signature\x18\x08 \x01(\x0cH\x00\x88\x01\x01\x42\x10\n\x0e_txn_signature\"\x8f\x01\n\x13SubscribeUpdateSlot\x12\x0c\n\x04slot\x18\x01 \x01(\x04\x12\x13\n\x06parent\x18\x02 \x01(\x04H\x00\x88\x01\x01\x12\"\n\x06status\x18\x03 \x01(\x0e\x32\x12.geyser.SlotStatus\x12\x17\n\ndead_error\x18\x04 \x01(\tH\x01\x88\x01\x01\x42\t\n\x07_parentB\r\n\x0b_dead_error\"g\n\x1aSubscribeUpdateTransaction\x12;\n\x0btransaction\x18\x01 \x01(\x0b\x32&.geyser.SubscribeUpdateTransactionInfo\x12\x0c\n\x04slot\x18\x02 \x01(\x04\"\xd8\x01\n\x1eSubscribeUpdateTransactionInfo\x12\x11\n\tsignature\x18\x01 \x01(\x0c\x12\x0f\n\x07is_vote\x18\x02 \x01(\x08\x12?\n\x0btransaction\x18\x03 \x01(\x0b\x32*.solana.storage.ConfirmedBlock.Transaction\x12\x42\n\x04meta\x18\x04 \x01(\x0b\x32\x34.solana.storage.ConfirmedBlock.TransactionStatusMeta\x12\r\n\x05index\x18\x05 \x01(\x04\"\xa1\x01\n SubscribeUpdateTransactionStatus\x12\x0c\n\x04slot\x18\x01 \x01(\x04\x12\x11\n\tsignature\x18\x02 \x01(\x0c\x12\x0f\n\x07is_vote\x18\x03 \x01(\x08\x12\r\n\x05index\x18\x04 \x01(\x04\x12<\n\x03\x65rr\x18\x05 \x01(\x0b\x32/.solana.storage.ConfirmedBlock.TransactionError\"\xa0\x04\n\x14SubscribeUpdateBlock\x12\x0c\n\x04slot\x18\x01 \x01(\x04\x12\x11\n\tblockhash\x18\x02 \x01(\t\x12\x37\n\x07rewards\x18\x03 \x01(\x0b\x32&.solana.storage.ConfirmedBlock.Rewards\x12@\n\nblock_time\x18\x04 \x01(\x0b\x32,.solana.storage.ConfirmedBlock.UnixTimestamp\x12@\n\x0c\x62lock_height\x18\x05 \x01(\x0b\x32*.solana.storage.ConfirmedBlock.BlockHeight\x12\x13\n\x0bparent_slot\x18\x07 \x01(\x04\x12\x18\n\x10parent_blockhash\x18\x08 \x01(\t\x12\"\n\x1a\x65xecuted_transaction_count\x18\t \x01(\x04\x12<\n\x0ctransactions\x18\x06 \x03(\x0b\x32&.geyser.SubscribeUpdateTransactionInfo\x12\x1d\n\x15updated_account_count\x18\n \x01(\x04\x12\x34\n\x08\x61\x63\x63ounts\x18\x0b \x03(\x0b\x32\".geyser.SubscribeUpdateAccountInfo\x12\x15\n\rentries_count\x18\x0c \x01(\x04\x12-\n\x07\x65ntries\x18\r \x03(\x0b\x32\x1c.geyser.SubscribeUpdateEntry\"\xe2\x02\n\x18SubscribeUpdateBlockMeta\x12\x0c\n\x04slot\x18\x01 \x01(\x04\x12\x11\n\tblockhash\x18\x02 \x01(\t\x12\x37\n\x07rewards\x18\x03 \x01(\x0b\x32&.solana.storage.ConfirmedBlock.Rewards\x12@\n\nblock_time\x18\x04 \x01(\x0b\x32,.solana.storage.ConfirmedBlock.UnixTimestamp\x12@\n\x0c\x62lock_height\x18\x05 \x01(\x0b\x32*.solana.storage.ConfirmedBlock.BlockHeight\x12\x13\n\x0bparent_slot\x18\x06 \x01(\x04\x12\x18\n\x10parent_blockhash\x18\x07 \x01(\t\x12\"\n\x1a\x65xecuted_transaction_count\x18\x08 \x01(\x04\x12\x15\n\rentries_count\x18\t \x01(\x04\"\x9d\x01\n\x14SubscribeUpdateEntry\x12\x0c\n\x04slot\x18\x01 \x01(\x04\x12\r\n\x05index\x18\x02 \x01(\x04\x12\x12\n\nnum_hashes\x18\x03 \x01(\x04\x12\x0c\n\x04hash\x18\x04 \x01(\x0c\x12\"\n\x1a\x65xecuted_transaction_count\x18\x05 \x01(\x04\x12\"\n\x1astarting_transaction_index\x18\x06 \x01(\x04\"\x15\n\x13SubscribeUpdatePing\"!\n\x13SubscribeUpdatePong\x12\n\n\x02id\x18\x01 \x01(\x05\"\x1c\n\x1aSubscribeReplayInfoRequest\"O\n\x1bSubscribeReplayInfoResponse\x12\x1c\n\x0f\x66irst_available\x18\x01 \x01(\x04H\x00\x88\x01\x01\x42\x12\n\x10_first_available\"\x1c\n\x0bPingRequest\x12\r\n\x05\x63ount\x18\x01 \x01(\x05\"\x1d\n\x0cPongResponse\x12\r\n\x05\x63ount\x18\x01 \x01(\x05\"\\\n\x19GetLatestBlockhashRequest\x12\x30\n\ncommitment\x18\x01 \x01(\x0e\x32\x17.geyser.CommitmentLevelH\x00\x88\x01\x01\x42\r\n\x0b_commitment\"^\n\x1aGetLatestBlockhashResponse\x12\x0c\n\x04slot\x18\x01 \x01(\x04\x12\x11\n\tblockhash\x18\x02 \x01(\t\x12\x1f\n\x17last_valid_block_height\x18\x03 \x01(\x04\"X\n\x15GetBlockHeightRequest\x12\x30\n\ncommitment\x18\x01 \x01(\x0e\x32\x17.geyser.CommitmentLevelH\x00\x88\x01\x01\x42\r\n\x0b_commitment\".\n\x16GetBlockHeightResponse\x12\x14\n\x0c\x62lock_height\x18\x01 \x01(\x04\"Q\n\x0eGetSlotRequest\x12\x30\n\ncommitment\x18\x01 \x01(\x0e\x32\x17.geyser.CommitmentLevelH\x00\x88\x01\x01\x42\r\n\x0b_commitment\"\x1f\n\x0fGetSlotResponse\x12\x0c\n\x04slot\x18\x01 \x01(\x04\"\x13\n\x11GetVersionRequest\"%\n\x12GetVersionResponse\x12\x0f\n\x07version\x18\x01 \x01(\t\"m\n\x17IsBlockhashValidRequest\x12\x11\n\tblockhash\x18\x01 \x01(\t\x12\x30\n\ncommitment\x18\x02 \x01(\x0e\x32\x17.geyser.CommitmentLevelH\x00\x88\x01\x01\x42\r\n\x0b_commitment\"7\n\x18IsBlockhashValidResponse\x12\x0c\n\x04slot\x18\x01 \x01(\x04\x12\r\n\x05valid\x18\x02 \x01(\x08*>\n\x0f\x43ommitmentLevel\x12\r\n\tPROCESSED\x10\x00\x12\r\n\tCONFIRMED\x10\x01\x12\r\n\tFINALIZED\x10\x02*\xa1\x01\n\nSlotStatus\x12\x12\n\x0eSLOT_PROCESSED\x10\x00\x12\x12\n\x0eSLOT_CONFIRMED\x10\x01\x12\x12\n\x0eSLOT_FINALIZED\x10\x02\x12\x1d\n\x19SLOT_FIRST_SHRED_RECEIVED\x10\x03\x12\x12\n\x0eSLOT_COMPLETED\x10\x04\x12\x15\n\x11SLOT_CREATED_BANK\x10\x05\x12\r\n\tSLOT_DEAD\x10\x06\x32\xf5\x04\n\x06Geyser\x12\x44\n\tSubscribe\x12\x18.geyser.SubscribeRequest\x1a\x17.geyser.SubscribeUpdate\"\x00(\x01\x30\x01\x12`\n\x13SubscribeReplayInfo\x12\".geyser.SubscribeReplayInfoRequest\x1a#.geyser.SubscribeReplayInfoResponse\"\x00\x12\x33\n\x04Ping\x12\x13.geyser.PingRequest\x1a\x14.geyser.PongResponse\"\x00\x12]\n\x12GetLatestBlockhash\x12!.geyser.GetLatestBlockhashRequest\x1a\".geyser.GetLatestBlockhashResponse\"\x00\x12Q\n\x0eGetBlockHeight\x12\x1d.geyser.GetBlockHeightRequest\x1a\x1e.geyser.GetBlockHeightResponse\"\x00\x12<\n\x07GetSlot\x12\x16.geyser.GetSlotRequest\x1a\x17.geyser.GetSlotResponse\"\x00\x12W\n\x10IsBlockhashValid\x12\x1f.geyser.IsBlockhashValidRequest\x1a .geyser.IsBlockhashValidResponse\"\x00\x12\x45\n\nGetVersion\x12\x19.geyser.GetVersionRequest\x1a\x1a.geyser.GetVersionResponse\"\x00\x42;Z9github.com/rpcpool/yellowstone-grpc/examples/golang/protoP\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'geyser_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'Z9github.com/rpcpool/yellowstone-grpc/examples/golang/proto'
  _globals['_SUBSCRIBEREQUEST_ACCOUNTSENTRY']._loaded_options = None
  _globals['_SUBSCRIBEREQUEST_ACCOUNTSENTRY']._serialized_options = b'8\001'
  _globals['_SUBSCRIBEREQUEST_SLOTSENTRY']._loaded_options = None
  _globals['_SUBSCRIBEREQUEST_SLOTSENTRY']._serialized_options = b'8\001'
  _globals['_SUBSCRIBEREQUEST_TRANSACTIONSENTRY']._loaded_options = None
  _globals['_SUBSCRIBEREQUEST_TRANSACTIONSENTRY']._serialized_options = b'8\001'
  _globals['_SUBSCRIBEREQUEST_TRANSACTIONSSTATUSENTRY']._loaded_options = None
  _globals['_SUBSCRIBEREQUEST_TRANSACTIONSSTATUSENTRY']._serialized_options = b'8\001'
  _globals['_SUBSCRIBEREQUEST_BLOCKSENTRY']._loaded_options = None
  _globals['_SUBSCRIBEREQUEST_BLOCKSENTRY']._serialized_options = b'8\001'
  _globals['_SUBSCRIBEREQUEST_BLOCKSMETAENTRY']._loaded_options = None
  _globals['_SUBSCRIBEREQUEST_BLOCKSMETAENTRY']._serialized_options = b'8\001'
  _globals['_SUBSCRIBEREQUEST_ENTRYENTRY']._loaded_options = None
  _globals['_SUBSCRIBEREQUEST_ENTRYENTRY']._serialized_options = b'8\001'
  _globals['_COMMITMENTLEVEL']._serialized_start=6299
  _globals['_COMMITMENTLEVEL']._serialized_end=6361
  _globals['_SLOTSTATUS']._serialized_start=6364
  _globals['_SLOTSTATUS']._serialized_end=6525
  _globals['_SUBSCRIBEREQUEST']._serialized_start=80
  _globals['_SUBSCRIBEREQUEST']._serialized_end=1388
  _globals['_SUBSCRIBEREQUEST_ACCOUNTSENTRY']._serialized_start=719
  _globals['_SUBSCRIBEREQUEST_ACCOUNTSENTRY']._serialized_end=806
  _globals['_SUBSCRIBEREQUEST_SLOTSENTRY']._serialized_start=808
  _globals['_SUBSCRIBEREQUEST_SLOTSENTRY']._serialized_end=889
  _globals['_SUBSCRIBEREQUEST_TRANSACTIONSENTRY']._serialized_start=891
  _globals['_SUBSCRIBEREQUEST_TRANSACTIONSENTRY']._serialized_end=986
  _globals['_SUBSCRIBEREQUEST_TRANSACTIONSSTATUSENTRY']._serialized_start=988
  _globals['_SUBSCRIBEREQUEST_TRANSACTIONSSTATUSENTRY']._serialized_end=1089
  _globals['_SUBSCRIBEREQUEST_BLOCKSENTRY']._serialized_start=1091
  _globals['_SUBSCRIBEREQUEST_BLOCKSENTRY']._serialized_end=1174
  _globals['_SUBSCRIBEREQUEST_BLOCKSMETAENTRY']._serialized_start=1176
  _globals['_SUBSCRIBEREQUEST_BLOCKSMETAENTRY']._serialized_end=1267
  _globals['_SUBSCRIBEREQUEST_ENTRYENTRY']._serialized_start=1269
  _globals['_SUBSCRIBEREQUEST_ENTRYENTRY']._serialized_end=1350
  _globals['_SUBSCRIBEREQUESTFILTERACCOUNTS']._serialized_start=1391
  _globals['_SUBSCRIBEREQUESTFILTERACCOUNTS']._serialized_end=1582
  _globals['_SUBSCRIBEREQUESTFILTERACCOUNTSFILTER']._serialized_start=1585
  _globals['_SUBSCRIBEREQUESTFILTERACCOUNTSFILTER']._serialized_end=1828
  _globals['_SUBSCRIBEREQUESTFILTERACCOUNTSFILTERMEMCMP']._serialized_start=1830
  _globals['_SUBSCRIBEREQUESTFILTERACCOUNTSFILTERMEMCMP']._serialized_end=1951
  _globals['_SUBSCRIBEREQUESTFILTERACCOUNTSFILTERLAMPORTS']._serialized_start=1953
  _globals['_SUBSCRIBEREQUESTFILTERACCOUNTSFILTERLAMPORTS']._serialized_end=2062
  _globals['_SUBSCRIBEREQUESTFILTERSLOTS']._serialized_start=2065
  _globals['_SUBSCRIBEREQUESTFILTERSLOTS']._serialized_end=2208
  _globals['_SUBSCRIBEREQUESTFILTERTRANSACTIONS']._serialized_start=2211
  _globals['_SUBSCRIBEREQUESTFILTERTRANSACTIONS']._serialized_end=2421
  _globals['_SUBSCRIBEREQUESTFILTERBLOCKS']._serialized_start=2424
  _globals['_SUBSCRIBEREQUESTFILTERBLOCKS']._serialized_end=2641
  _globals['_SUBSCRIBEREQUESTFILTERBLOCKSMETA']._serialized_start=2643
  _globals['_SUBSCRIBEREQUESTFILTERBLOCKSMETA']._serialized_end=2677
  _globals['_SUBSCRIBEREQUESTFILTERENTRY']._serialized_start=2679
  _globals['_SUBSCRIBEREQUESTFILTERENTRY']._serialized_end=2708
  _globals['_SUBSCRIBEREQUESTACCOUNTSDATASLICE']._serialized_start=2710
  _globals['_SUBSCRIBEREQUESTACCOUNTSDATASLICE']._serialized_end=2777
  _globals['_SUBSCRIBEREQUESTPING']._serialized_start=2779
  _globals['_SUBSCRIBEREQUESTPING']._serialized_end=2813
  _globals['_SUBSCRIBEUPDATE']._serialized_start=2816
  _globals['_SUBSCRIBEUPDATE']._serialized_end=3381
  _globals['_SUBSCRIBEUPDATEACCOUNT']._serialized_start=3383
  _globals['_SUBSCRIBEUPDATEACCOUNT']._serialized_end=3494
  _globals['_SUBSCRIBEUPDATEACCOUNTINFO']._serialized_start=3497
  _globals['_SUBSCRIBEUPDATEACCOUNTINFO']._serialized_end=3697
  _globals['_SUBSCRIBEUPDATESLOT']._serialized_start=3700
  _globals['_SUBSCRIBEUPDATESLOT']._serialized_end=3843
  _globals['_SUBSCRIBEUPDATETRANSACTION']._serialized_start=3845
  _globals['_SUBSCRIBEUPDATETRANSACTION']._serialized_end=3948
  _globals['_SUBSCRIBEUPDATETRANSACTIONINFO']._serialized_start=3951
  _globals['_SUBSCRIBEUPDATETRANSACTIONINFO']._serialized_end=4167
  _globals['_SUBSCRIBEUPDATETRANSACTIONSTATUS']._serialized_start=4170
  _globals['_SUBSCRIBEUPDATETRANSACTIONSTATUS']._serialized_end=4331
  _globals['_SUBSCRIBEUPDATEBLOCK']._serialized_start=4334
  _globals['_SUBSCRIBEUPDATEBLOCK']._serialized_end=4878
  _globals['_SUBSCRIBEUPDATEBLOCKMETA']._serialized_start=4881
  _globals['_SUBSCRIBEUPDATEBLOCKMETA']._serialized_end=5235
  _globals['_SUBSCRIBEUPDATEENTRY']._serialized_start=5238
  _globals['_SUBSCRIBEUPDATEENTRY']._serialized_end=5395
  _globals['_SUBSCRIBEUPDATEPING']._serialized_start=5397
  _globals['_SUBSCRIBEUPDATEPING']._serialized_end=5418
  _globals['_SUBSCRIBEUPDATEPONG']._serialized_start=5420
  _globals['_SUBSCRIBEUPDATEPONG']._serialized_end=5453
  _globals['_SUBSCRIBEREPLAYINFOREQUEST']._serialized_start=5455
  _globals['_SUBSCRIBEREPLAYINFOREQUEST']._serialized_end=5483
  _globals['_SUBSCRIBEREPLAYINFORESPONSE']._serialized_start=5485
  _globals['_SUBSCRIBEREPLAYINFORESPONSE']._serialized_end=5564
  _globals['_PINGREQUEST']._serialized_start=5566
  _globals['_PINGREQUEST']._serialized_end=5594
  _globals['_PONGRESPONSE']._serialized_start=5596
  _globals['_PONGRESPONSE']._serialized_end=5625
  _globals['_GETLATESTBLOCKHASHREQUEST']._serialized_start=5627
  _globals['_GETLATESTBLOCKHASHREQUEST']._serialized_end=5719
  _globals['_GETLATESTBLOCKHASHRESPONSE']._serialized_start=5721
  _globals['_GETLATESTBLOCKHASHRESPONSE']._serialized_end=5815
  _globals['_GETBLOCKHEIGHTREQUEST']._serialized_start=5817
  _globals['_GETBLOCKHEIGHTREQUEST']._serialized_end=5905
  _globals['_GETBLOCKHEIGHTRESPONSE']._serialized_start=5907
  _globals['_GETBLOCKHEIGHTRESPONSE']._serialized_end=5953
  _globals['_GETSLOTREQUEST']._serialized_start=5955
  _globals['_GETSLOTREQUEST']._serialized_end=6036
  _globals['_GETSLOTRESPONSE']._serialized_start=6038
  _globals['_GETSLOTRESPONSE']._serialized_end=6069
  _globals['_GETVERSIONREQUEST']._serialized_start=6071
  _globals['_GETVERSIONREQUEST']._serialized_end=6090
  _globals['_GETVERSIONRESPONSE']._serialized_start=6092
  _globals['_GETVERSIONRESPONSE']._serialized_end=6129
  _globals['_ISBLOCKHASHVALIDREQUEST']._serialized_start=6131
  _globals['_ISBLOCKHASHVALIDREQUEST']._serialized_end=6240
  _globals['_ISBLOCKHASHVALIDRESPONSE']._serialized_start=6242
  _globals['_ISBLOCKHASHVALIDRESPONSE']._serialized_end=6297
  _globals['_GEYSER']._serialized_start=6528
  _globals['_GEYSER']._serialized_end=7157
# @@protoc_insertion_point(module_scope)
